import logging

import numpy as np
import scipy.sparse.linalg as la
from numpy.linalg import norm

from aspire.abinitio import CLOrient3D
from aspire.operators import PolarFT
from aspire.utils import J_conjugate, Rotation, all_pairs, all_triplets, tqdm, trange
from aspire.utils.random import randn
from aspire.volume import DnSymmetryGroup

logger = logging.getLogger(__name__)


class CLSymmetryD2(CLOrient3D):
    """
    Define a class to estimate 3D orientations using common lines methods for
    molecules with D2 (dihedral) symmetry.

    Corresponding publication:
    E. Rosen and Y. Shkolnisky,
    Common lines ab-initio reconstruction of D2-symmetric molecules,
    SIAM Journal on Imaging Sciences, volume 13-4, p. 1898-1994, 2020
    """

    def __init__(
        self,
        src,
        n_rad=None,
        n_theta=None,
        max_shift=0.15,
        shift_step=1,
        grid_res=1200,
        inplane_res=5,
        eq_min_dist=7,
        epsilon=0.01,
        seed=None,
        mask=True,
    ):
        """
        Initialize object for estimating 3D orientations for molecules with D2 symmetry.

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param n_rad: The number of points in the radial direction of Fourier image.
        :param n_theta: The number of points in the theta direction of Fourier image.
        :param max_shift: Maximum range for shifts as a proportion of resolution. Default = 0.15.
        :param shift_step: Resolution of shift estimation in pixels. Default = 1 pixel.
        :param grid_res: Number of sampling points on sphere for projetion directions.
            These are generated using the Saaf-Kuijlaars algorithm. Default value is 1200.
        :param inplane_res: The sampling resolution of in-plane rotations for each
            projection direction. Default value is 5 degrees.
        :param eq_min_dist: Width of strip around equator projection directions from
            which we do not sample directions. Default value is 7 degrees.
        :param epsilon: Tolerance for J-synchronization power method.
        :param seed: Optional seed for RNG.
        :param mask: Option to mask `src.images` with a fuzzy mask (boolean).
            Default, `True`, applies a mask.
        """

        super().__init__(
            src,
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
            mask=mask,
        )

        self.grid_res = grid_res
        self.inplane_res = inplane_res
        self.n_inplane_rots = int(360 / self.inplane_res)
        self.eq_min_dist = eq_min_dist
        self.seed = seed
        self.epsilon = epsilon

        self.triplets = all_triplets(self.n_img)
        self.pairs, self.pairs_to_linear = all_pairs(self.n_img, return_map=True)
        self.n_pairs = len(self.pairs)

        # D2 symmetry group.
        # Rearrange in order Identity, about_x, about_y, about_z.
        # This ordering is necessary for reproducing MATLAB code results.
        self.gs = DnSymmetryGroup(order=2, dtype=self.dtype).matrices[[0, 3, 2, 1]]

    def estimate_rotations(self):
        """
        Estimate rotation matrices for molecules with D2 symmetry. Sets the attribute
        self.rotations with an array of estimated rotation matrices, size src.nx3x3.
        """
        # Pre-compute phase-shifted polar Fourier.
        self._compute_shifted_pf()

        # Generate lookup data
        self._generate_lookup_data()
        self._generate_scl_lookup_data()

        # Compute self common-line scores.
        self._compute_scl_scores()

        # Compute common-lines and estimate relative rotations Rijs.
        self._compute_cl_scores()

        # Perform handedness synchronization.
        self.Rijs_sync = self._global_J_sync(self.Rijs_est)

        # Synchronize colors.
        self.colors, self.Rijs_rows = self._sync_colors(self.Rijs_sync)

        # Synchronize signs.
        Ris = self._sync_signs(self.Rijs_rows, self.colors)

        # Assign rotations.
        self.rotations = Ris

    #########################
    # Prepare Polar Fourier #
    #########################

    def _compute_shifted_pf(self):
        """
        Pre-compute shifted and full polar Fourier transforms.
        """
        logger.info("Preparing polar Fourier transform.")
        pf = self.pf

        # Generate shift phases.
        r_max = pf.shape[-1]
        max_shift_1d = np.ceil(2 * np.sqrt(2) * self.max_shift)
        shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
            r_max, max_shift_1d, self.shift_step
        )
        self.n_shifts = len(shifts)

        # Reconstruct full polar Fourier for use in correlation.
        pf[:, :, 0] = 0  # Matching matlab convention to zero out the lowest frequency.
        pf /= norm(pf, axis=2)[..., np.newaxis]  # Normalize each ray.
        self.pf_full = PolarFT.half_to_full(pf)

        # Pre-compute shifted pf's.
        pf_shifted = pf[:, None] * shift_phases[None, :, None]
        self.pf_shifted = pf_shifted.reshape(
            (self.n_img, self.n_shifts * (self.n_theta // 2), r_max)
        )

    ###################################
    # Generate Commonline Lookup Data #
    ###################################

    def _generate_lookup_data(self):
        """
        Generate candidate relative rotations and corresponding common line indices.
        """
        logger.info("Generating commonline lookup data.")
        # Generate uniform grid on sphere with Saff-Kuijlaars and take one quarter
        # of sphere because of D2 symmetry redundancy.
        sphere_grid = self._saff_kuijlaars(self.grid_res)
        octant1_mask = np.all(sphere_grid > 0, axis=1)
        octant2_mask = (
            (sphere_grid[:, 0] > 0) & (sphere_grid[:, 1] > 0) & (sphere_grid[:, 2] < 0)
        )
        sphere_grid1 = sphere_grid[octant1_mask]
        sphere_grid2 = sphere_grid[octant2_mask]

        # Mark Equator Directions.
        # Common lines between projection directions which are perpendicular to
        # symmetry axes (equator images) have common line degeneracies. Two images
        # taken from directions on the same great circle which is perpendicular to
        # some symmetry axis only have 2 common lines instead of 4, and must be
        # treated separately.
        # We detect such directions by taking a strip of radius
        # `eq_min_dist` about the 3 great circles perpendicular to the symmetry
        # axes of D2 (i.e to X,Y and Z axes).
        eq_class1 = self._mark_equators(sphere_grid1, self.eq_min_dist)
        eq_class2 = self._mark_equators(sphere_grid2, self.eq_min_dist)

        #  Mark Top View Directions.
        #  A Top view projection image is taken from the direction of one of the
        #  symmetry axes. Since all symmetry axes of D2 molecules are perpendicular
        #  this means that such an image is an equator with repect to both symmetry
        #  axes which are perpendicular to the direction of the symmetry axis from
        #  which the image was made, e.g. if the image was formed by projecting in
        #  the direction of the X (symmetry) axis, then it is an equator with
        #  respect to both Y and Z symmetry axes (it's direction is the
        #  interesection of 2 great circles perpendicular to Y and Z axes).
        #  Such images have severe degeneracies. A pair of Top View images (taken
        #  from different directions or a Top View and equator image only have a
        #  single common line. A top view and a regular non-equator image only have
        #  two common lines.

        # Remove top views from sphere grids and update equator indices and classes.
        self.sphere_grid1 = sphere_grid1[eq_class1 < 4]
        self.sphere_grid2 = sphere_grid2[eq_class2 < 4]
        self.eq_class1 = eq_class1[eq_class1 < 4]
        self.eq_class2 = eq_class2[eq_class2 < 4]

        # Generate in-plane rotations for each grid point on the sphere.
        self.inplane_rotated_grid1 = self._generate_inplane_rots(
            self.sphere_grid1, self.inplane_res
        )
        self.inplane_rotated_grid2 = self._generate_inplane_rots(
            self.sphere_grid2, self.inplane_res
        )

        # Generate commmonline angles induced by all relative rotation candidates.
        cl_angles1, self.eq2eq_Rij_table_11 = self._generate_commonline_angles(
            self.inplane_rotated_grid1,
            self.inplane_rotated_grid1,
            self.eq_class1,
            self.eq_class1,
        )
        cl_angles2, self.eq2eq_Rij_table_12 = self._generate_commonline_angles(
            self.inplane_rotated_grid1,
            self.inplane_rotated_grid2,
            self.eq_class1,
            self.eq_class2,
            same_octant=False,
        )

        # Generate commonline indices.
        self.cl_idx_1 = self._generate_commonline_indices(cl_angles1)
        self.cl_idx_2 = self._generate_commonline_indices(cl_angles2)
        self.cl_idx = np.hstack((self.cl_idx_1, self.cl_idx_2))

    def _generate_commonline_angles(
        self,
        Ris,
        Rjs,
        Ri_eq_class,
        Rj_eq_class,
        same_octant=True,
    ):
        """
        Compute commonline angles induced by the 4 sets of relative rotations
        Rij = Ri.T @ g_m @ Rj, m = 0,1,2,3, where g_m is the identity and rotations
        about the three axes of symmetry of a D2 symmetric molecule. Note, we only
        compute commonline angles between pairs of images which are not equator
        images with respect to the same axis of symmetry. To do this we build a
        table, `eq2eq_Rij_table`, which is `False` for pairs of images that are
        equator images with respect to the same axis of symmetry and `True` otherwise.

        :param Ris: First set of candidate rotations.
        :param Rjs: Second set of candidate rotation.
        :param Ri_eq_class: Equator classification for Ris.
        :param Rj_eq_class: Equator classification for Rjs.
        :param same_octant: True if both sets of candidates are in the same octant.

        :return: Commonline angles induced by relative rotation candidates.
        """
        n_rots_i = len(Ris)
        n_theta = Ris.shape[1]  # Same for Rjs, TODO: Don't call this n_theta

        # Generate upper triangular table of indicators of all pairs which are not
        # equators with respect to the same symmetry axis (named unique_pairs).
        eq_table = np.outer(Ri_eq_class > 0, Rj_eq_class > 0)
        in_same_class = (Ri_eq_class[:, None] - Rj_eq_class.T[None]) == 0
        eq2eq_Rij_table = ~(eq_table * in_same_class)

        # For candidates in the same octant only need upper triangle of table.
        if same_octant:
            eq2eq_Rij_table = np.triu(eq2eq_Rij_table, 1)

        n_pairs = np.count_nonzero(eq2eq_Rij_table)
        idx = 0
        cl_angles = np.zeros((2, n_pairs, n_theta, n_theta // 2, 4, 2))

        for i in range(n_rots_i):
            unique_pairs_i = np.nonzero(eq2eq_Rij_table[i])[0]
            if len(unique_pairs_i) == 0:
                continue
            Ri = Ris[i]
            for j in unique_pairs_i:
                Rj = Rjs[j, : n_theta // 2]

                # Compute relative rotations candidates Rij = Ri.T @ gs @ Rj
                Rijs = (
                    np.transpose(Ri, axes=(0, 2, 1))[:, None, None]
                    @ self.gs
                    @ Rj[:, None]
                )

                # Common line indices induced by Rijs
                cl_angles[0, idx, :, :, :, 0] = np.arctan2(
                    -Rijs[..., 0, 2], Rijs[..., 1, 2]
                )
                cl_angles[0, idx, :, :, :, 1] = np.arctan2(
                    Rijs[..., 2, 0], -Rijs[..., 2, 1]
                )
                cl_angles[1, idx, :, :, :, 0] = np.arctan2(
                    -Rijs[..., 2, 0], Rijs[..., 2, 1]
                )
                cl_angles[1, idx, :, :, :, 1] = np.arctan2(
                    Rijs[..., 0, 2], -Rijs[..., 1, 2]
                )

                idx += 1

        # Make all angles non-negative and convert to degrees.
        cl_angles = (cl_angles + 2 * np.pi) % (2 * np.pi)
        cl_angles = cl_angles * 180 / np.pi

        return cl_angles, eq2eq_Rij_table

    ########################################
    # Generate Self-Commonline Lookup Data #
    ########################################

    def _generate_scl_lookup_data(self):
        """
        Generate lookup data for self-commonlines.
        """
        logger.info("Generating self-commonline lookup data.")
        # Get self-commonline angles.
        self.scl_angles1 = self._generate_scl_angles(
            self.inplane_rotated_grid1,
            self.eq_class1,
        )
        self.scl_angles2 = self._generate_scl_angles(
            self.inplane_rotated_grid2,
            self.eq_class2,
        )

        # Get self-commonline indices.
        self.scl_idx_1, self.scl_eq_lin_idx_lists_1 = self._generate_scl_indices(
            self.scl_angles1, self.eq_class1
        )
        self.scl_idx_2, self.scl_eq_lin_idx_lists_2 = self._generate_scl_indices(
            self.scl_angles2, self.eq_class2
        )
        self.scl_idx_lists = np.concatenate(
            (self.scl_eq_lin_idx_lists_1, self.scl_eq_lin_idx_lists_2), axis=1
        )

        # Compute non-equator indices.
        # Register non equator indices. Denote by C_ij the j'th in-plane rotation of
        # the i'th ML candidate, and arrange all candidates in a list with their in-plane
        # rotations in the order: C_11,...,C_1r,...,C_m1,...,C_mr where m is the
        # number of candidates and r is the number of in plane rotations. Here we
        # create a sub-list of only non equator candidates, i.e., if i_1,...,i_p are
        # non equators then we have the sub list is
        # C_(i_1)1,...,C(i_1)r,...C_(i_p)1,...,C_(i_p)r.
        n_non_eq = np.count_nonzero(self.eq_class1 == 0) + np.count_nonzero(
            self.eq_class2 == 0
        )
        non_eq_idx = np.zeros((n_non_eq, self.n_inplane_rots), dtype=int)
        non_eq_idx[:, 0] = (
            np.hstack(
                (
                    np.nonzero(self.eq_class1 == 0)[0],
                    len(self.eq_class1) + np.nonzero(self.eq_class2 == 0)[0],
                )
            )
            * self.n_inplane_rots
        )
        non_eq_idx[:, 1:] = non_eq_idx[:, [0]] + np.arange(1, self.n_inplane_rots)

        self.non_eq_idx = non_eq_idx

        # Non-topview equator indices.
        self.non_tv_eq_idx = np.concatenate(
            (
                np.nonzero(self.eq_class1 > 0)[0],
                len(self.eq_class1) + np.nonzero(self.eq_class2 > 0)[0],
            )
        )

        # Generate maps from scl indices to relative rotations.
        self._generate_scl_scores_idx_map()

    def _generate_scl_angles(self, Ris, eq_class):
        """
        Generate self-commonline angles. For each candidate rotation a pair of self-commonline
        angles are generated for each of the 3 self-commonlines induced by D2 symmetry.

        :param Ris: Candidate rotation matrices, (n_sphere_grid, n_inplane_rots, 3, 3).
        :param eq_idx: Equator index mask for Ris.
        :param eq_class: Equator classification for Ris.

        :return: `scl_angles` of shape (n_sphere_grid, n_inplane_rots, 3, 2).
        """

        # For each candidate rotation Ri we generate the set of 3 self-commonlines.
        scl_angles = np.zeros((*Ris.shape[:2], 3, 2), dtype=Ris.dtype)
        n_rots = len(Ris)
        for i in range(n_rots):
            Ri = Ris[i]
            for k, g in enumerate(self.gs[1:]):
                g_Ri = g @ Ri
                Riis = np.transpose(Ri, axes=(0, 2, 1)) @ g_Ri

                scl_angles[i, :, k, 0] = np.arctan2(Riis[:, 2, 0], -Riis[:, 2, 1])
                scl_angles[i, :, k, 1] = np.arctan2(-Riis[:, 0, 2], Riis[:, 1, 2])

        # Prepare self commonline coordinates.
        scl_angles = scl_angles % (2 * np.pi)

        # Deal with non top view equators
        # A non-TV equator has only one self common line. However, we clasify an
        # equator as an image whose projection direction is at radial distance <
        # `eq_min_dist` from the great circle perpendicular to a symmetry axis,
        # and not strictly zero distance. Thus in most cases we get 2 common lines
        # differing by a small difference in degrees. Actually the calculation above
        # gives us two NEARLY antipodal lines, so we first flip one of them by
        # adding 180 degrees to it. Then we aggregate all the rays within the range
        # between these two resulting lines to compute the score of this self common
        # line for this candidate. The scoring part is done in the ML function itself.
        # Furthermore, the line perpendicular to the self common line, though not
        # really a self common line, has the property that all its values are real
        # and both halves of the line (rays differing by pi, emanating from the
        # origin) have the same values, and so it 'behaves like' a self common
        # line which we also register here and exploit in the ML function.
        # We put the 'real' self common line at 2 first coordinates, the
        # candidate for perpendicular line is in 3rd coordinate.

        # If this is a self common line with respect to x-equator then the actual self
        # common line(s) is given by the self relative rotations given by the y and z
        # rotation (by 180 degrees) group members, i.e. Ri^TgyRj and Ri^TgzRj
        scl_angles[eq_class == 1] = scl_angles[eq_class == 1][:, :, [1, 2, 0]]
        scl_angles[eq_class == 1, :, 0] = scl_angles[eq_class == 1][:, :, 0, [1, 0]]

        # If this is a self common line with respect to y-equator then the actual self
        # common line(s) is given by the self relative rotations given by the x and z
        # rotation (by 180 degrees) group members, i.e. Ri^TgxRj and Ri^TgzRj
        scl_angles[eq_class == 2] = scl_angles[eq_class == 2][:, :, [0, 2, 1]]
        scl_angles[eq_class == 2, :, 0] = scl_angles[eq_class == 2][:, :, 0, [1, 0]]

        # If this is a self common line with respect to z-equator then the actual self
        # common line(s) is given by the self relative rotations given by the x and y
        # rotation (by 180 degrees) group members, i.e. Ri^TgxRj and Ri^TgyRj
        # No need to rearrange entries, the "real" common lines are already in
        # indices 1 and 2, but flip one common line to antipodal.
        scl_angles[eq_class == 3, :, 0] = scl_angles[eq_class == 3][:, :, 0, [1, 0]]

        # Make sure angle range is < 180 degrees.
        # p1 marks "equator" self-commonlines where both entries of the first
        # scl are greater than both entries of the second scl.
        p1 = scl_angles[eq_class > 0, :, 0] > scl_angles[eq_class > 0, :, 1]
        p1 = p1[:, :, 0] & p1[:, :, 1]
        # p2 marks "equator" self-commonlines where the angle range between the
        # first and second sets of self-commonlines is greater than 180.
        p2 = scl_angles[eq_class > 0, :, 0] - scl_angles[eq_class > 0, :, 1] < -np.pi
        p2 = p2[:, :, 0] | p2[:, :, 1]
        p = p1 | p2

        # Swap entries satisfying either of the above conditions.
        scl_angles[eq_class > 0] = (
            scl_angles[eq_class > 0][:, :, [1, 0, 2]] * p[:, :, None, None]
            + scl_angles[eq_class > 0] * ~p[:, :, None, None]
        )

        # Convert from radians [0,2*pi) to degrees [0, 360).
        return np.round(scl_angles * 180 / np.pi) % 360

    def _generate_scl_indices(self, scl_angles, eq_class):
        """
        Generate self-commonline indices. This includes a set of linear indices for
        all candidate rotations as well as lists of self-commonline index ranges for
        equator candidates.

        :param scl_angles: Self-commonline angles, shape (n_sphere_grid, n_inplane_rots, 3, 2).
        :param eq_class: Equator classification for the sphere_grid points represented
            by the first axis of `scl_angles`.

        :returns:
            - scl_indices, self-commonline linear indices.
            - eq_lin_idx_lists, a list containing a range of self-commonline
                indices for each equator candidate.
        """
        L = self.n_theta

        # Convert from angles to indices.
        scl_indices = self._generate_commonline_indices(scl_angles)
        scl_angles = np.mod(np.round(scl_angles / (2 * np.pi) * L), L).astype(int)

        # Create candidate common line linear indices lists for equators.
        # As indicated above for equator candidate, for each self common line we
        # don't get a single coordinate but a range of them. Here we register a
        # list of coordinates for each such self common line candidate.
        non_top_view_eq_idx = np.nonzero(eq_class > 0)[0]
        n_eq = len(non_top_view_eq_idx)
        n_inplane_rots = scl_angles.shape[1]
        count_eq = 0

        # eq_lin_idx_lists[0,i,j] registers a list of linear indices of the j'th
        # in-plane rotation of the range for the (only) self common line of the i'th
        # candidate. eq_lin_idx_lists[1,i,j] registers the actual (integer) angle
        # of the self common line in the 2D Fourier space. Note that we need only
        # one number since each self common line has radial coordinates of the form
        # (theta, theta+180).
        eq_lin_idx_lists = np.empty((2, n_eq, n_inplane_rots), dtype=object)
        for i in non_top_view_eq_idx.tolist():
            for j in range(n_inplane_rots):
                idx1 = self._circ_seq(scl_angles[i, j, 0, 0], scl_angles[i, j, 1, 0], L)
                idx2 = self._circ_seq(scl_angles[i, j, 0, 1], scl_angles[i, j, 1, 1], L)

                # Ensure idx1 and idx2 have same number of elements.
                # Might be off by one due to n_theta discretization.
                end = np.minimum(len(idx1), len(idx2))
                idx1, idx2 = idx1[:end], idx2[:end]

                # Adjust so idx1 is in [0, 180) range.
                is_geq_than_pi = idx1 >= L // 2
                idx1[is_geq_than_pi] = idx1[is_geq_than_pi] - L // 2
                idx2[is_geq_than_pi] = (idx2[is_geq_than_pi] + L // 2) % L

                # register indices in list.
                eq_lin_idx_lists[0, count_eq, j] = np.ravel_multi_index(
                    (idx1, idx2), (L // 2, L)
                )
                eq_lin_idx_lists[1, count_eq, j] = idx1
            count_eq += 1

        return scl_indices, eq_lin_idx_lists

    def _generate_scl_scores_idx_map(self):
        """
        Generates lookup tables for maximum likelihood scheme to estimate commonlines
        between images.

        This method creates two lookup tables (`oct1_ij_map` and `oct2_ij_map`)
        for pairs of candidate rotations (i, j) under the following conditions:

        1. Both rotations Ri and Rj are in octant 1.
        2. Ri is in octant 1 and Rj is in octant 2.

        For each pair of candidate rotations the tables give a map into the set of
        self-commonlines induced by those rotations. This table will be used later
        to incorporate a likelihood score for self-commonlines into the likelihood
        score for common lines for each pair of images.
        """
        # Calculate number of rotations in each octant.
        n_rot_1 = len(self.scl_idx_1) // (3 * self.n_inplane_rots)
        n_rot_2 = len(self.scl_idx_2) // (3 * self.n_inplane_rots)

        # First the map for i<j pairs for the rotations Ri and Rj in octant 1
        # which are not equator images with respect to the same axis of symmetry.
        n_pairs = np.count_nonzero(self.eq2eq_Rij_table_11)
        oct1_ij_map = np.zeros(
            (n_pairs, self.n_inplane_rots**2 // 2, 2), dtype=np.int64
        )

        # Create index arrays for i and j to cover all rotation combinations.
        i_idx = np.repeat(np.arange(self.n_inplane_rots), self.n_inplane_rots // 2)
        j_idx = np.tile(np.arange(self.n_inplane_rots // 2), self.n_inplane_rots)

        idx_vec = np.arange(n_rot_1)
        idx = 0

        for i in range(n_rot_1):
            unique_pairs_i = idx_vec[self.eq2eq_Rij_table_11[i]]
            if len(unique_pairs_i) == 0:
                continue
            i_idx_plus_offset = i_idx + (i * self.n_inplane_rots)

            for j in unique_pairs_i:
                j_idx_plus_offset = j_idx + (j * self.n_inplane_rots)
                oct1_ij_map[idx] = np.column_stack(
                    (i_idx_plus_offset, j_idx_plus_offset)
                )
                idx += 1

        # Now the map for i<j pairs for Ri in octant 1 and Rj in octant 2.
        n_pairs_12 = np.count_nonzero(self.eq2eq_Rij_table_12)
        oct2_ij_map = np.zeros(
            (n_pairs_12, self.n_inplane_rots**2 // 2, 2), dtype=np.int64
        )
        idx_vec = np.arange(n_rot_2)
        idx = 0

        for i in range(n_rot_1):
            unique_pairs_i = idx_vec[self.eq2eq_Rij_table_12[i] > 0]
            if len(unique_pairs_i) == 0:
                continue
            i_idx_plus_offset = i_idx + (i * self.n_inplane_rots)

            for j in unique_pairs_i:
                j_idx_plus_offset = j_idx + (j * self.n_inplane_rots)
                oct2_ij_map[idx] = np.column_stack(
                    (i_idx_plus_offset, j_idx_plus_offset)
                )
                idx += 1

        tmp1 = oct1_ij_map[:, :, 0].flatten()
        tmp2 = oct1_ij_map[:, :, 1].flatten()
        self.oct1_ij_map = np.column_stack((tmp1, tmp2))

        tmp1 = oct2_ij_map[:, :, 0].flatten()
        tmp2 = oct2_ij_map[:, :, 1].flatten()
        self.oct2_ij_map = np.column_stack((tmp1, tmp2))

    ##############################################
    # Compute Self-Commonline Correlation Scores #
    ##############################################

    def _compute_scl_scores(self):
        """
        Compute correlations for self-commonline candidates. For each image i
        we compute an auto-correlation table between all polar Fourier rays.
        We then use that table to apply a score to each non-topview candidate
        rotation which gives the likelihood that the self-commonlines induced
        by that candidate belong to the image i..
        """
        logger.info("Computing self-commonline correlation scores.")
        n_img = self.n_img
        n_theta = self.n_theta
        n_eq = len(self.non_tv_eq_idx)
        n_inplane = self.n_inplane_rots

        # Prepare self-commonline indices.
        scl_matrix = np.concatenate((self.scl_idx_1, self.scl_idx_2))
        M = len(scl_matrix) // 3
        scl_idx = scl_matrix.reshape(M, 3)

        # Get non-equator indices to use with corrs matrix.
        non_eq_lin_idx = self.non_eq_idx.flatten()
        n_non_eq = len(non_eq_lin_idx)
        non_eq_idx = np.unravel_index(
            scl_idx[non_eq_lin_idx].flatten(), (n_theta // 2, n_theta)
        )

        # Compute max correlation over all shifts.
        corrs = np.real(
            self.pf_shifted @ np.transpose(np.conj(self.pf_full), (0, 2, 1))
        )
        corrs = np.reshape(corrs, (self.n_img, self.n_shifts, n_theta // 2, n_theta))
        corrs = np.max(corrs, axis=1)

        # Map correlations to probabilities (in the spirit of Maximum Likelihood).
        corrs = 0.5 * (corrs + 1)

        # Compute equator measures.
        eq_measures = np.zeros((self.n_img, n_theta // 2), dtype=self.dtype)
        for i in range(self.n_img):
            eq_measures[i] = self._all_eq_measures(corrs[i])

        # Handle the cases: Non-equator, Non-top-view equator images.
        # 1. Non-equators: just take product of probabilities.
        corrs_out = np.zeros((n_img, M), dtype=self.dtype)
        prod_corrs = np.prod(
            corrs[:, non_eq_idx[0], non_eq_idx[1]].reshape(self.n_img, n_non_eq, 3),
            axis=2,
        )
        corrs_out[:, non_eq_lin_idx] = prod_corrs

        # 2. Non-topview equators: adjust scores by eq_measures
        for eq_idx in range(n_eq):
            for j in range(n_inplane):
                # Take the correlations for the self common line candidate of the
                # "equator rotation" `eq_idx` with respect to image i, and
                # multiply by all scores from the function eq_measures (see
                # documentation inside the function ). Then take maximum over
                # all the scores.
                scl_idx_list = np.unravel_index(
                    self.scl_idx_lists[0, eq_idx, j], (n_theta // 2, n_theta)
                )
                true_scls_corrs = corrs[:, scl_idx_list[0], scl_idx_list[1]]
                scls_cand_idx = self.scl_idx_lists[1, eq_idx, j]
                eq_measures_j = eq_measures[:, scls_cand_idx]
                measures_agg = true_scls_corrs[:, :, None] * eq_measures_j[:, None, :]
                k = self.non_tv_eq_idx[eq_idx]
                corrs_out[:, k * n_inplane + j] = np.max(measures_agg, axis=(-2, -1))

        self.scls_scores = corrs_out

    def _all_eq_measures(self, corrs):
        """
        Compute a measure indicating how likely an image is an equator image.

        :param corrs: Correlation table of shape (n_theta // 2, n_theta).

        :return: (n_theta // 2) likelihood scores.
        """
        # First compute the eq measure (corrs(scl-k,scl+k) for k=1:n_theta // 4)
        # An equator image of a D2 molecule has the following property: If t_i is
        # the angle of one of the rays of the self common line then all the pairs of
        # rays of the form (t_i-k,t_i+k) for k=1:n_theta // 4 are identical. For each t_i we
        # average over correlations between the lines (t_i-k,t_i+k) for k=1:n_theta // 4
        # to measure the likelihood that the image is an equator and the ray (line)
        # with angle t_i is a self common line.
        # (This first loop can be done once outside this function and then pass
        # idx as an argument).
        L = self.n_theta
        L_half = L // 2

        # Generate indices using broadcasting.
        t_i = np.arange(L_half)[:, None, None]
        k_vals = np.arange(1, L // 4 + 1)[None, :, None]
        neg_pos_k = np.array([-1, 1])[None, None, :]

        # Calculate indices, shape: (L//2, L//4, 2).
        idx = np.mod(t_i + k_vals * neg_pos_k, L)

        # Convert to Fourier ray indices.
        idx_1 = idx[:, :, 0].flatten()
        idx_2 = idx[:, :, 1].flatten()

        # Adjust idx_1 to be within [0, 180) and adjust idx_2 accordingly.
        is_geq_than_pi = idx_1 >= L_half
        idx_1[is_geq_than_pi] -= L_half
        idx_2[is_geq_than_pi] = (idx_2[is_geq_than_pi] + L_half) % L

        # Compute correlations
        eq_corrs = corrs[idx_1, idx_2].reshape(L_half, L // 4)
        corrs_mean = np.mean(eq_corrs, axis=1)

        # Now compute correlations for normals to scls.
        # An eqautor image of a D2 molecule has the additional following property:
        # The normal line to a self common line in 2D Fourier plane is real valued
        # and both of its rays have identical values. We use the correlation
        # between one Fourier ray of the normal to a self common line candidate t_i
        # with its anti-podal as an additional way to measure if the image is an
        # equator and t_i+0.5*pi is the normal to its self common line.
        r = np.ceil(2 * L / 360).astype(
            int
        )  # Search radius within 2 degrees of normal ray.

        # Generate indices for normal to scl index.
        normal_2_scl_idx_0 = (
            L_half - np.arange(L_half // 2 - r, L_half // 2 + r + 1)
        ) % L
        normal_2_scl_idx = (normal_2_scl_idx_0 + np.arange(L_half).reshape(-1, 1)) % L

        # Adjust indices to be within [0, 180) range.
        normal_2_scl_idx = np.where(
            normal_2_scl_idx >= L_half, normal_2_scl_idx - L_half, normal_2_scl_idx
        )

        # Compute correlations for normals.
        normal_corrs = corrs[normal_2_scl_idx, normal_2_scl_idx + L_half]
        normal_corrs_max = np.max(normal_corrs, axis=1)

        return corrs_mean * normal_corrs_max

    #########################################
    # Compute Commonline Correlation Scores #
    #########################################

    def _compute_cl_scores(self):
        """
        Run common lines Maximum likelihood procedure for a D2 molecule, to find
        the set of rotations Ri^TgkRj, k=1,2,3,4 for each pair of images i and j.
        """
        logger.info("Computing commonline correlation scores.")
        L = self.n_theta
        n_pairs = self.n_img * (self.n_img - 1) // 2

        # Map the self common line scores of each 2 candidate rotations R_i, R_j
        n_lookup_1 = len(self.scl_idx_1) // 3
        oct1_ij_map = np.vstack((self.oct1_ij_map, self.oct1_ij_map[:, [1, 0]]))
        oct2_ij_map = self.oct2_ij_map
        oct2_ij_map[:, 1] += n_lookup_1
        oct2_ij_map = np.vstack((oct2_ij_map, oct2_ij_map[:, [1, 0]]))
        ij_map = np.vstack((oct1_ij_map, oct2_ij_map))

        # Gather commonline indices and unravel to index into correlations.
        cl_idx = np.unravel_index(self.cl_idx, (L // 2, L))

        # Allocate output variables
        corrs_idx = np.zeros(n_pairs, dtype=np.int64)
        corrs_out = np.zeros(n_pairs, dtype=self.dtype)

        ij_idx = 0
        pbar = tqdm(
            desc="Searching for commonlines between pairs of images", total=n_pairs
        )

        # For each i'th image compute the correlation with all j'th images, j > i.
        for i in range(self.n_img - 1):
            pf_i = self.pf_shifted[i]
            scores_i = self.scls_scores[i]

            # Gather all pf_j in one array for vectorized computation
            pf_js = self.pf_full[i + 1 : self.n_img]
            n_pf_js = pf_js.shape[0]

            # Compute maximum correlation over all shifts for all pf_j
            corrs = np.real(pf_i @ np.conj(pf_js.transpose(0, 2, 1)))
            corrs = corrs.reshape(n_pf_js, self.n_shifts, L // 2, L)
            corrs = np.max(corrs, axis=1)  # Max over shifts

            # Take the product over symmetrically induced candidates. Eq. 4.5 in paper.
            prod_corrs = corrs[:, cl_idx[0], cl_idx[1]]
            prod_corrs = prod_corrs.reshape(n_pf_js, len(prod_corrs[0]) // 4, 4)
            prod_corrs = np.prod(prod_corrs, axis=2)

            # Incorporate scores of individual rotations from self-commonlines
            scores_js = self.scls_scores[i + 1 : self.n_img]
            scores_ij = scores_i[ij_map[:, 0]] * scores_js[:, ij_map[:, 1]]

            # Find maximum correlations and update results
            prod_corrs = prod_corrs * scores_ij
            max_indices = np.argmax(prod_corrs, axis=1)
            corrs_idx[ij_idx : ij_idx + len(max_indices)] = max_indices
            corrs_out[ij_idx : ij_idx + len(max_indices)] = prod_corrs[
                np.arange(len(max_indices)), max_indices
            ]

            ij_idx += len(max_indices)
            pbar.update(len(max_indices))

        pbar.close()

        # Get estimated relative viewing directions
        self.corrs_idx = corrs_idx
        self.Rijs_est = self._get_Rijs_from_lin_idx(corrs_idx)

    def _get_Rijs_from_lin_idx(self, lin_idx):
        """
        Restore map results from maximum-likelihood over commonlines to corresponding
        relative rotations.

        :param lin_idx: Set of linear indices corresponding to best estimate of Rijs.

        :return: Estimated Rijs.
        """
        Rijs_est = np.zeros((len(lin_idx), 4, 3, 3), dtype=self.dtype)
        n_cand_per_oct = len(self.cl_idx_1) // 4
        oct1_idx = lin_idx < n_cand_per_oct
        n_est_in_oct1 = np.count_nonzero(oct1_idx)
        if n_est_in_oct1 > 0:
            Rijs_est[oct1_idx] = self._get_Rijs_from_oct(lin_idx[oct1_idx], octant=1)
        if n_est_in_oct1 <= len(lin_idx):
            Rijs_est[~oct1_idx] = self._get_Rijs_from_oct(
                lin_idx[~oct1_idx] - n_cand_per_oct, octant=2
            )

        return Rijs_est

    def _get_Rijs_from_oct(self, lin_idx, octant=1):
        """
        Calculate estimated relative rotations Rijs from the linear indices of
        common-lines estimates from the search table. Rijs are generated from the
        rotation grids from which the common-lines table was generated.

        :param lin_idx: Set of linear indices corresponding to best estimate of Rijs.
        :param octant: Octant of rotation grid from which the Rj rotation was selected
             when generating the common-lines table.
        :return: Estimated Rijs.
        """
        if octant not in [1, 2]:
            raise ValueError("`octant` must be 1 or 2.")

        # Get pairs lookup table.
        if octant == 1:
            unique_pairs = self.eq2eq_Rij_table_11
        else:
            unique_pairs = self.eq2eq_Rij_table_12

        n_theta = self.n_inplane_rots
        n_lookup_pairs = np.count_nonzero(unique_pairs)
        n_rots = len(self.sphere_grid1)
        if octant == 1:
            n_rots2 = n_rots
        else:
            n_rots2 = len(self.sphere_grid2)

        # Map linear indices of chosen pairs of rotation candidates from ML to regular indices.
        p_idx, inplane_i, inplane_j = np.unravel_index(
            lin_idx, (2 * n_lookup_pairs, n_theta, n_theta // 2)
        )
        transpose_idx = p_idx >= n_lookup_pairs
        p_idx[transpose_idx] -= n_lookup_pairs
        s = self.inplane_rotated_grid1.shape
        inplane_rotated_grid = np.reshape(
            self.inplane_rotated_grid1, (np.prod(s[0:2]), 3, 3)
        )
        if octant == 1:
            s2 = s
            inplane_rotated_grid2 = inplane_rotated_grid
        else:
            s2 = self.inplane_rotated_grid2.shape
            inplane_rotated_grid2 = np.reshape(
                self.inplane_rotated_grid2, (np.prod(s2[0:2]), 3, 3)
            )

        # Convert linear indices of unique table to linear indices of index pairs table.
        idx_vec = np.arange(np.prod(unique_pairs.shape))
        unique_lin_idx = idx_vec[unique_pairs.flatten()]
        I, J = np.unravel_index(unique_lin_idx, (n_rots, n_rots2))
        est_idx = np.vstack((I[p_idx], J[p_idx]))

        # Assemble relative rotations Ri^TgRj using linear indices, where g is a group member of D2.
        Ris_lin_idx = np.ravel_multi_index((est_idx[0], inplane_i), s[:2])
        Rjs_lin_idx = np.ravel_multi_index((est_idx[1], inplane_j), s2[:2])
        Ris_t = np.transpose(inplane_rotated_grid[Ris_lin_idx], (0, 2, 1))
        Rjs = inplane_rotated_grid2[Rjs_lin_idx]
        Rijs_est = Ris_t[:, None] @ self.gs @ Rjs[:, None]

        Rijs_est[transpose_idx] = np.transpose(Rijs_est[transpose_idx], (0, 1, 3, 2))

        return Rijs_est

    ####################################
    # Perform Global J Synchronization #
    ####################################

    def _global_J_sync(self, Rijs):
        """
        Global J-synchronization of all third row outer products. Given n_pairsx4x3x3
        matrices Rijs, each of which might contain a spurious J, ie.
        Rij = J @ Ri.T @ gs @ Rj @ J instead of Rij = Ri.T @ gs @ Rj, we return Rijs
        that all have either a spurious J or not.

        :param Rijs: An (n-choose-2)x4 x3x3 array where each 3x3 slice holds an estimate
            for the corresponding outer-product Ri.T @ Rj. Each estimate might have a
            spurious J independently of other estimates.

        :return: Rijs, all of which have a spurious J or not.
        """
        logger.info("Performing global handedness synchronization.")
        # Find best J_configuration.
        J_list = self._J_configuration(Rijs)

        # Determine relative handedness of Rijs.
        sign_ij_J = self._J_sync_power_method(J_list)

        # Synchronize Rijs
        logger.info("Applying global handedness synchronization.")
        mask = sign_ij_J == 1
        Rijs[mask] = J_conjugate(Rijs[mask])

        return Rijs

    def _J_configuration(self, Rijs):
        """
        For each triplet of indices (i, j, k), consider the relative rotations
        tuples {Ri^TgmRj}, {Ri^TglRk} and {Rj^TgrRk}. Compute norms of the form
        ||Ri^TgmRj*Rj^TglRk-Ri^TglRk||, ||J*Ri^TgmRj*J*Rj^TglRk-Ri^TglRk||,
        ||Ri^TgmRj*J*Rj^TglRk*J-Ri^TglRk| and ||Ri^TgmRj*Rj^TglRk-J*Ri^TglRk*J||
        where gm,gl,gr are the varipus gorup members of Dn and J=diag([1,1-1]).
        The correct "J-configuration" is given for the smallest of these 4 norms.

        :param Rijs: (n-choose-2)x3x3 array of relative rotations.
        :return: List of n-choose-3 indices in {0,1,2,3} indicating
            which J-configuration for each triplet of Rijs, i<j<k.
        """
        J_list = np.zeros(len(self.triplets), dtype="int")
        R = Rijs
        trip_idx = 0
        pbar = tqdm(
            desc="Finding best J-configuration over triplets", total=len(self.triplets)
        )
        for i, j, k in self.triplets:
            ij = self.pairs_to_linear[i, j]
            jk = self.pairs_to_linear[j, k]
            ik = self.pairs_to_linear[i, k]
            Rij, Rjk, Rik = R[ij], R[jk], R[ik]
            Rjk_t = np.transpose(Rjk, (0, 2, 1))

            Rij_J = J_conjugate(Rij)
            Rjk_t_J = J_conjugate(Rjk_t)
            Rik_J = J_conjugate(Rik)

            final_votes = np.zeros(4)
            final_votes[0] = self._compare_rots(Rij, Rjk_t, Rik)
            final_votes[1] = self._compare_rots(Rij_J, Rjk_t, Rik)
            final_votes[2] = self._compare_rots(Rij, Rjk_t_J, Rik)
            final_votes[3] = self._compare_rots(Rij, Rjk_t, Rik_J)

            J_list[trip_idx] = np.argmin(final_votes)
            trip_idx += 1

            pbar.update()
        pbar.close()

        return J_list

    def _compare_rots(self, Rij, Rjk_t, Rik):
        """
        Compute norms for the 4 J-configurations and return indices
        corresponding to best configuration for the provided triplet
        of relative rotations.

        :param Rij: Relative rotation between i'th and j'th candidate rotations
            of shape (4, 3, 3).
        :param Rjk_t: Transpose of relative rotation between j'th and k'th candidate
            rotations of shape (4, 3, 3).
        :param Rik: Relative rotation between i'th and k'th candidate rotations
            of shape (4, 3, 3).
        :return: Score for this J-configuration of the given rotation triplet.
        """
        # We compute the four sets of 4^3 norms |Rik @ Rjk.T - Rij|
        # See equation (6.11) in publication.
        prod_arr = Rik[:, None] @ Rjk_t[None]
        diff_arr = prod_arr[:, :, None] - Rij
        diff_arr = diff_arr.reshape((64, 9))
        norm_arr = np.sum(diff_arr**2, axis=1)

        # For perfect estimates, 16 of the 64 norms will equal zero.
        # We sum over the smallest 16 values to get a vote for this J-configuration.
        m = np.sort(norm_arr)
        vote = np.sum(m[:16])

        return vote

    def _J_sync_power_method(self, J_list):
        """
        Calculate the leading eigenvector of the J-synchronization matrix
        using the power method.

        As the J-synchronization matrix is of size (n-choose-2)x(n-choose-2), we
        use the power method to compute the eigenvalues and eigenvectors,
        while constructing the matrix on-the-fly.

        :param Rijs: (n-choose-2)x3x3 array of estimates of relative orientation matrices.

        :return: An array of length n-choose-2 consisting of 1 or -1, where the sign
            of the i'th entry indicates whether the i'th relative orientation matrix
            will be J-conjugated.
        """

        # Set power method tolerance and maximum iterations.
        epsilon = self.epsilon
        max_iters = 100

        # Initialize candidate eigenvectors
        vec = randn(self.n_pairs, seed=self.seed)
        vec = vec / norm(vec)
        residual = 1
        itr = 0

        # Power method iterations
        logger.info(
            "Initiating power method to estimate J-synchronization matrix eigenvector."
        )
        while itr < max_iters and residual > epsilon:
            itr += 1
            vec_new = self._signs_times_v(J_list, vec)
            vec_new = vec_new / norm(vec_new)
            residual = norm(vec_new - vec)
            vec = vec_new
            logger.info(
                f"Iteration {itr}, residual {round(residual, 5)} (target {epsilon})"
            )

        # We need only the signs of the eigenvector
        J_sync = np.sign(vec)
        J_sync = np.sign(J_sync[0]) * J_sync  # Stabilize J_sync

        return J_sync

    def _signs_times_v(self, J_list, vec):
        """
        Multiplication of the J-synchronization matrix by a candidate eigenvector.

        The J-synchronization matrix is a matrix representation of the handedness graph,
        Gamma, whose set of nodes consists of the estimates Rijs and whose set of edges
        consists of the undirected edges between all triplets of estimates Rij, Rjk,
        and Rik, where i<j<k. The weight of an edge is set to +1 if its incident nodes
        agree in handednes and -1 if not.

        The J-synchronization matrix is of size (n-choose-2)x(n-choose-2), where each
        entry corresponds to the relative handedness of Rij and Rjk. The entry (ij, jk),
        where ij and jk are retrieved from the all_pairs indexing, is 1 if Rij and Rjk
        are of the same handedness and -1 if not. All other entries (ij, kl) hold a zero.

        Due to the large size of the J-synchronization matrix we construct it on the fly
        as follows. For each triplet of outer products Rij, Rjk, and Rik, the associated
        elements of the J-synchronization matrix are populated with +1 or -1 and
        multiplied by the corresponding elements of the current candidate eigenvector
        supplied by the power method. The new candidate eigenvector is updated for each
        triplet.

        :param J_list: n-choose-3 array of indices indicating the best signs configuration.
        :param vec: The current candidate eigenvector of length n-choose-2 from the power
            method.

        :return: New candidate eigenvector of length n-choose-2. The product of the J-sync
            matrix and vec.
        """
        new_vec = np.zeros_like(vec)
        signs_confs = np.array(
            [[1, 1, 1], [-1, 1, -1], [-1, -1, 1], [1, -1, -1]], dtype=int
        )
        trip_idx = 0
        for i in trange(self.n_img, desc="Computing signs_times_v"):
            for j in range(i + 1, self.n_img - 1):
                ij = self.pairs_to_linear[i, j]
                for k in range(j + 1, self.n_img):
                    ik = self.pairs_to_linear[i, k]
                    jk = self.pairs_to_linear[j, k]

                    best_i = J_list[trip_idx]
                    trip_idx += 1

                    s_ij_jk = signs_confs[best_i][0]
                    s_ik_jk = signs_confs[best_i][1]
                    s_ij_ik = signs_confs[best_i][2]

                    # Update multiplication
                    new_vec[ij] += s_ij_jk * vec[jk] + s_ij_ik * vec[ik]
                    new_vec[jk] += s_ij_jk * vec[ij] + s_ik_jk * vec[ik]
                    new_vec[ik] += s_ij_ik * vec[ij] + s_ik_jk * vec[jk]

        return new_vec

    ######################
    # Synchronize Colors #
    ######################

    def _sync_colors(self, Rijs):
        """
        At this point, we have obtained a hand-consistent set of 4-tuples of Rijs,
        with Rij = Ri.T @ g_s @ Rj, where s is an unknown permutation of (0, 1, 2, 3).
        Taking the average of the first Rij in the 4-tuple with the remaining 3
        results in a set of 3 outer products (+-)vi(m).T @ vj(m) of unknown ordering
        where vi(m) is the m'th row of the i'th rotation matrix, Ri.

        The color sync procedure partitions the set of 3-tuples of m'th row outer
        products into 3 sets of row-consistent outer products up to the sign of each.

        :param Rijs: Array of shape (n_pairs,4,3,3) consisting of the n_pairs of
            hand-consistent 4-tuples of Rijs.
        :returns:
            - cp, A color mapping vector of length (n_pairs * 3) which permutes
                the 3-tuples of `Rijs_rows` to be globally row-consistent.
            - Rijs_rows, An array of color synchronized rotations' rows outer products of
                shape (n_pairs, 3, 3, 3), where each Rijs_rows[ij] corresponds to a 3-tuple
                of m'th row outer product matrices, some of which having a spurious -1.
        """
        logger.info("Performing rotations' rows synchronization.")
        # Generate array of one rank matrices from which we can extract rows.
        # Matrices are of the form 0.5(Ri^TRj+Ri^TgkRj). Each such matrix can be
        # written in the form Qi^T*Ik*Qj where Ik is a 3x3 matrix with all zero
        # entries except for the entry a_kk, k in {1,2,3}.
        n_pairs = len(Rijs)
        Rijs_rows = np.zeros((n_pairs, 3, 3, 3), dtype=self.dtype)
        for layer in range(3):
            Rijs_rows[:, layer] = 0.5 * (Rijs[:, 0] + Rijs[:, layer + 1])

        # Partition the set of matrices Rijs_rows into 3 sets of matrices, where
        # each set there are only matrices Qi^T*Ik*Qj for a unique value of k in
        # {1,2,3}.
        # First determine for each pair of tuples of the form {Qi^T*Ik*Qj} and
        # {Qr^T*Il*Qj} where {i,j}\cap{r,l}==1, whether l==r.
        color_perms = self._match_colors(Rijs_rows)

        # Compute eigenvectors of color matrix. This is just a matrix of dimensions
        # 3(N choose 2)x3(N choose 2) where each entry corresponds to a pair of
        # matrices {Qi^T*Ir*Qj} and {Qr^T*Il*Qj} and equals \delta_rl.
        # The 2 leading eigenvectors span a linear subspace which contains a
        # vector which encodes the partition. All the entries of the vector are
        # either 1,0 or -1, where the number encodes which the index r in Ir.
        # This vector is a linear combination of the two leading eigen vectors,
        # and so we 'unmix' these vectors to retrieve it.
        color_mat = la.LinearOperator(
            (3 * n_pairs,) * 2, lambda v: self._mult_cmat_by_vec(color_perms, v)
        )

        # Seed eigs initial vector for iterative method.
        # scipy LinearOperator needs doubles for some architectures (arm).
        v0 = randn(3 * n_pairs, seed=self.seed).astype(np.float64, copy=False)

        v0 = v0 / norm(v0)
        vals, colors = la.eigs(color_mat, k=3, which="LR", v0=v0)
        vals = np.real(vals)
        colors = np.real(colors).astype(self.dtype, copy=False)
        colors = np.sign(colors[0]) * colors  # Stable eigs
        cp, _ = self._unmix_colors(colors[:, :2])

        return cp, Rijs_rows

    def _match_colors(self, Rijs_rows):
        """
        For each triplet of indices i < j < k, we consider the m'th row outer products stored
        as Rijs_rows, ie. Rijs_rows[ij], Rijs_rows[jk], and Rijs_rows[ik]. Recall that
        Rijs_rows[ij, n], n=0,1,2, corresponds to the 3x3 outer product vi_m.T @ vj_m, where
        vi_m is an unknown row of the rotation matrices Ri and Rj. For each triplet of these
        sets of row outer products this method finds a permutation sigma such that
        Rijs_rows[ij, sigma(n)], Rijs_rows[jk, sigma(n)], and Rijs_rows[ik, sigma(n)] all
        correspond to the same m'th row outer product.

        Framed as graph partioning problem we are coloring the vertices, Rijs_rows[ij, n],
        with three colors such that each color corresponds to the same row of the rotations
        Ris. This method returns the permutation that rearanges the elements of each triplet
        of Rijs to have matching color.

        :param Rijs_rows: An n_pairsx3x3x3 array of m'th row outer products for the pairs
            Ri, Rj, where Rijs_rows[:, i] is the m'th row outer product of unknown row m.
        :return: n_pairs length array corresponding to the permutation which color matches
            Rijs_rows.
        """
        Rijs_rows_t = np.transpose(Rijs_rows, (0, 1, 3, 2))
        trip_perms = np.array(
            [[0, 1, 2], [0, 2, 1], [1, 0, 2], [1, 2, 0], [2, 0, 1], [2, 1, 0]],
            dtype="int",
        )
        inverse_perms = np.array(
            [
                [0, 1, 2],
                [0, 2, 1],
                [1, 0, 2],
                [2, 0, 1],
                [1, 2, 0],
                [2, 1, 0],
            ],
            dtype="int",
        )

        m = np.zeros((6, 6), dtype=self.dtype)
        colors_i = np.zeros((len(self.triplets), 3), dtype=int)
        n_trip = len(self.triplets)
        votes = np.zeros((n_trip))
        trip_idx = 0

        # Compute relative color permutations. See Section 7.2 of paper.
        for i, j, k in self.triplets:
            ij = self.pairs_to_linear[i, j]
            jk = self.pairs_to_linear[j, k]
            ik = self.pairs_to_linear[i, k]

            # For r=1:3 compute 3*3 products v_{ji}(r)v_{ik}v_{kj}
            prod_arr = Rijs_rows[ik, None] @ Rijs_rows_t[jk, :, None]
            prod_arr_tmp = prod_arr.copy()
            prod_arr = Rijs_rows_t[ij, :, None] @ prod_arr_tmp.reshape((9, 3, 3))[None]
            prod_arr = np.transpose(
                prod_arr.reshape((3, 3, 3, 9), order="F"), (2, 1, 0, 3)
            )

            # Compare to v_{jj}(r)=v_{ji}v_{ij}.
            self_prods = Rijs_rows_t[ij] @ Rijs_rows[ij]
            self_prods = self_prods.reshape(3, 9)

            prod_arr1 = prod_arr.copy()
            prod_arr1 -= self_prods
            norms1 = np.sum(prod_arr1**2, axis=3)

            prod_arr2 = prod_arr.copy()
            prod_arr2 += self_prods
            norms2 = np.sum(prod_arr2**2, axis=3)

            # Compare to v_{jj}(r)=v_{jk}v_{kj}.
            self_prods = Rijs_rows[jk] @ Rijs_rows_t[jk]
            self_prods = self_prods.reshape(3, 9)

            prod_arr1 = prod_arr.copy()
            prod_arr1 -= self_prods[:, None, None]
            norms1 = norms1 + np.sum(prod_arr1**2, axis=3)

            prod_arr2 = prod_arr.copy()
            prod_arr2 += self_prods[:, None, None]
            norms2 = norms2 + np.sum(prod_arr2**2, axis=3)

            # For r=1:3 compute 3*3 products v_{ij}(r)v_{jk}v_{ki} and compare to
            # Compare to v_{ii}(r)=v_{ij}v_{ji}
            prod_arr = np.transpose(prod_arr_tmp, (0, 1, 3, 2))
            prod_arr = Rijs_rows[ij, :, None] @ prod_arr.reshape((9, 3, 3))[None]
            prod_arr = np.transpose(
                prod_arr.reshape((3, 3, 3, 9), order="F"), (1, 0, 2, 3)
            )

            # Compare to v_{ii}(r)=v_{ik}v_{ki}.
            self_prods = Rijs_rows[ik] @ Rijs_rows_t[ik]
            self_prods = self_prods.reshape(3, 9)

            prod_arr1 = prod_arr.copy()
            prod_arr1 -= self_prods[None, :, None]
            norms1 = norms1 + np.sum(prod_arr1**2, axis=3)

            prod_arr2 = prod_arr.copy()
            prod_arr2 += self_prods[None, :, None]
            norms2 = norms2 + np.sum(prod_arr2**2, axis=3)

            norms = np.minimum(norms1, norms2)

            for r in range(6):
                p1 = trip_perms[r]
                for s in range(6):
                    p2 = trip_perms[s]
                    m[r, s] = (
                        norms[p2[0], p1[0], 0]
                        + norms[p2[1], p1[1], 1]
                        + norms[p2[2], p1[2], 2]
                    )

            # In the event of duplicate min values min_idx is the first occurence
            # by column order to match matlab outputs.
            min_idx = np.unravel_index(np.argmin(m.T), m.shape)[::-1]
            votes[trip_idx] = m[min_idx]

            # Store permutation indices as digits of a base 10 number.
            colors_i[trip_idx, :2] = [
                100 * (min_idx[0] + 1),
                10 * (min_idx[1] + 1),
            ]

            # Calculate the relative permutation of Rik to Rij given
            # by (sigma_ik)\circ(sigma_ij)^-1
            inv_jk_perm = inverse_perms[min_idx[1]]
            rel_perm = trip_perms[min_idx[0]]
            rel_perm = rel_perm[inv_jk_perm]
            colors_i[trip_idx, 2] = (2 * (rel_perm[0] + 1) - 1) + (
                rel_perm[1] > rel_perm[2]
            )
            trip_idx += 1

        colors_i = np.sum(colors_i, axis=1)

        return colors_i

    def _mult_cmat_by_vec(self, c_perms, v):
        """
        Multiply color matrix by vector v "on the fly".

        :param c_perms: An (N over 3) vector. Each corresponds to a triplet of
            indices i<j<k and indicates the relative permutation of tuples
            {Qi^TIrQj},{Qj^TIlQk} and {Qk^TIlQi}. The color matrix can be
            completely reconstructed from this information, which is used
            here to execute a single multiplication of the matrix by the
            vector v instead of explicitely computing and storing the prohibitively
            large color matrix in memory.
        :param v: vector to be multiplied by 'color matrix'.
        """
        t_perms = np.array(
            [[0, 1, 2], [0, 2, 1], [1, 0, 2], [1, 2, 0], [2, 0, 1], [2, 1, 0]]
        )
        i_perms = np.array(
            [[0, 1, 2], [0, 2, 1], [1, 0, 2], [2, 0, 1], [1, 2, 0], [2, 1, 0]]
        )
        out = np.zeros_like(v)
        trip_idx = 0
        for i in range(self.n_img):
            for j in range(i + 1, self.n_img - 1):
                ij_block = 3 * self.pairs_to_linear[i, j]
                for k in range(j + 1, self.n_img):
                    ik_block = 3 * self.pairs_to_linear[i, k]
                    jk_block = 3 * self.pairs_to_linear[j, k]

                    # Extract permutation indices from c_perms
                    n = c_perms[trip_idx]
                    trip_idx += 1
                    p_n1 = n // 100
                    p_n3 = n % 10
                    p_n2 = (n - p_n1 * 100 - p_n3) // 10

                    # Adjust for 0-based indexing. (Take this out by computing c_perms with 0-base)
                    p_n1 = (p_n1 - 1).astype("int")
                    p_n2 = (p_n2 - 1).astype("int")
                    p_n3 = (p_n3 - 1).astype("int")

                    # Multiply vector by color matrix

                    # Upper triangular part
                    p = t_perms[p_n1] + ik_block
                    out[ij_block] = out[ij_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[ij_block + 1] = out[ij_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[ij_block + 2] = out[ij_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]

                    p = t_perms[p_n2] + jk_block
                    out[ij_block] = out[ij_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[ij_block + 1] = out[ij_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[ij_block + 2] = out[ij_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]

                    p = i_perms[p_n3] + jk_block
                    out[ik_block] = out[ik_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[ik_block + 1] = out[ik_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[ik_block + 2] = out[ik_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]

                    # Lower triangular part
                    p = i_perms[p_n1] + ij_block
                    out[ik_block] = out[ik_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[ik_block + 1] = out[ik_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[ik_block + 2] = out[ik_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]

                    p = i_perms[p_n2] + ij_block
                    out[jk_block] = out[jk_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[jk_block + 1] = out[jk_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[jk_block + 2] = out[jk_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]

                    p = t_perms[p_n3] + ik_block
                    out[jk_block] = out[jk_block] - v[p[1]] - v[p[2]] + v[p[0]]
                    out[jk_block + 1] = out[jk_block + 1] - v[p[0]] - v[p[2]] + v[p[1]]
                    out[jk_block + 2] = out[jk_block + 2] - v[p[0]] - v[p[1]] + v[p[2]]
        return out

    def _unmix_colors(self, color_vecs):
        """
        The 'color vector' which partitions the rank 1 3x3 matrices into 3 sets
        is one of 2 leading orthogonal eigenvectors of the color matrix.
        SVD retrieves two orthogonal linear combinations of these vectors which
        can be 'unmixed' to retrieve the color vector by finding a suitable
        2D rotation of these vectors (see Section 7.3 of D2 paper for details).
        """
        n_p = color_vecs.shape[0] // 3
        d_theta = self.n_theta // self.n_theta
        max_t = self.n_theta // d_theta + 1

        def R_theta(theta):
            R = np.array(
                [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]],
                dtype=self.dtype,
            )
            return R

        s = float("inf")
        scores = np.zeros(max_t, dtype=self.dtype)
        idx = 0
        for t in np.arange(0, max_t, 0.5):
            unmix_ev = color_vecs @ R_theta(np.pi * t / 180)
            s1 = unmix_ev[:, 0].reshape(n_p, 3)
            p11 = (-s1).argsort(axis=1)  # descending argsort
            s1 = np.take_along_axis(s1, p11, axis=1)
            score11 = np.sum((s1[:, 0] + s1[:, 2]) ** 2 + s1[:, 1] ** 2)

            s2 = abs(unmix_ev[:, 1].reshape(n_p, 3))
            p12 = (-s2).argsort(axis=1)  # descending argsort
            s2 = np.take_along_axis(s2, p12, axis=1)
            score12 = np.sum(
                (s2[:, 0] - 2 * s2[:, 1]) ** 2
                + (s2[:, 0] - 2 * s2[:, 2]) ** 2
                + (s2[:, 1] - s2[:, 2]) ** 2
            )  # Matlab comment: Is this an error??? + instead of - in the first 2 members

            s1 = abs(unmix_ev[:, 0].reshape(n_p, 3))
            p12 = (-s1).argsort(axis=1)  # descending argsort
            s1 = np.take_along_axis(s1, p11, axis=1)
            score22 = np.sum(
                (s1[:, 0] - 2 * s1[:, 1]) ** 2
                + (s1[:, 0] - 2 * s1[:, 2]) ** 2
                + (s1[:, 1] - s1[:, 2]) ** 2
            )

            s2 = unmix_ev[:, 1].reshape(n_p, 3)
            p22 = (-s2).argsort(axis=1)  # descending argsort
            s2 = np.take_along_axis(s2, p12, axis=1)
            score21 = np.sum((s2[:, 0] + s2[:, 2]) ** 2 + s2[:, 1] ** 2)

            score_vecs = [score11 + score12, score21 + score22]
            which_vec = np.argmin([score11 + score12, score21 + score22])
            scores[idx] = score_vecs[which_vec]
            if scores[idx] < s:
                s = scores[idx]
                if which_vec == 0:
                    p = p11
                else:
                    p = p22
                best_unmix = unmix_ev[:, which_vec]

            # Assign integers between 1:3 to permutations
            colors = np.zeros((n_p, 3), dtype=int)
            for i in range(n_p):
                p_i = p[i]
                p_i_sqr = p_i[p_i]
                if np.sum((p_i_sqr - [0, 1, 2]) ** 2) == 0:  # non-cyclic permutation
                    colors[i] = p_i
                else:
                    colors[i] = p_i_sqr
            colors = colors.flatten()

        return colors, best_unmix

    #####################
    # Synchronize Signs #
    #####################

    def _sync_signs(self, rr, c_vec):
        """
        This function executes the final stage of the algorithm, Signs
        synchroniztion. At this point, we have rotation rows
        rr[ij, m] = sij_m * vi_m.T @ vj_m, where vi_m, vj_m are the m'th rows
        of rotation matrices Ri and Rj and sij_m is an unknown sign. This method
        uses the permutation vector, `c_vec`, to partition the rotation row
        outer products and constructs a symmetric block matrix, H, with ij'th block
        sij * vi.T @ vj. The signs sij are then adjusted so that H is rank-1. This
        matrix is then factored to extract the rows of each rotation matrix. At the
        end all rows of the rotations Ri are exctracted and the matrices Ri are assembled.

        :param rr: Array of color synchronized rotations' rows outer products of
            shape (n_pairs, 3, 3, 3), where each rr[ij] corresponds to a 3-tuple
            of m'th row outer product matrices, some of which having a spurious -1.
        :param c_vec: A color mapping vector of length (n_pairs * 3) which permutes
            the 3-tuples of `rr` to be globally row-consistent.
        :return: n_img x 3 x 3 array of rotation matrices.
        """
        logger.info("Performing signs synchronization.")
        c_mat, c_mat_5d, c_mat_4d = self._construct_color_mats(rr, c_vec)

        sync_signs2 = self._compute_signs(c_mat_5d, c_mat_4d)

        rows_arr = self._estimate_rows(sync_signs2, c_mat_5d)

        signs = self._compute_signs_adjustment(rows_arr)

        rots = self._extract_rotations(c_mat, signs)

        return rots

    def _construct_color_mats(self, rr, c_vec):
        """
        Construct the partitioned row synchronized color matrices, `c_mat`, where
        c_mat[m] contains the 3x3 blocks sij*vi_m.T @ vj_m, where vi_m is the m'th
        row of the i'th rotation Ri and sij is the unknown sign.

        :param rr: Non-partitioned rotation row matrices.
        :param c_vec: Color partition vector.
        :return: Partitioned row synchronized color matrices.
        """
        # Partition the union of tuples {0.5*(Ri^TRj+Ri^TgkRj), k=1:3} according
        # to the color partition established in color synchronization procedure.
        # The partition is stored in two different arrays each with the purpose
        # of a computational speed up for two different computations performed
        # later (space considerations are of little concern since arrays are ~
        # o(N^2) which doesn't pose a constraint for inputs on the scale of 10^3-10^4.
        c_mat_5d = np.zeros((self.n_img, self.n_img, 3, 3, 3), dtype=self.dtype)
        c_mat_4d = np.zeros((self.n_pairs, 3, 3, 3), dtype=self.dtype)
        c_vec = c_vec.reshape(self.n_pairs, 3)
        for i in range(self.n_img - 1):
            for j in range(i + 1, self.n_img):
                ij = self.pairs_to_linear[i, j]
                c_mat_5d[i, j, c_vec[ij]] = rr[ij]
                c_mat_5d[j, i, c_vec[ij]] = rr[ij].transpose(0, 2, 1)
                c_mat_4d[ij, c_vec[ij]] = rr[ij]

        # Compute estimates for the tuples {0.5*(Ri^TRi+Ri^TgkRi), k=1:3} for
        # i=1:N. For 1<=i,j<=N and c=1,2,3 write Qij^c=0.5*(Ri^TRj+Ri^TgmRj).
        # For each i in {1:N} and each k in {1,2,3} the estimator is the
        # average over all j~=i of Qij^c*(Qij^c)^T.
        # Since in practice the result of the average is not really rank 1, we
        # compute the best rank approximation to this average.
        for i in range(self.n_img):
            for c in range(3):
                Rijs = c_mat_5d[i, :, c]
                Rijs = np.delete(Rijs, i, axis=0)
                Rii_est = Rijs @ np.transpose(Rijs, (0, 2, 1))
                Rii = np.mean(Rii_est, axis=0)
                U, _, _ = np.linalg.svd(Rii)
                c_mat_5d[i, i, c] = np.outer(U[:, 0], U[:, 0])

        # Construct the 3Nx3N row synchroniztion matrices (as done for C_2), one
        # for all first rows of the matrices Ri, one for all second rows and one
        # for all third rows. The ij'th block of the k'th matrix is Qij^c.
        # In C_2 one such matrix is constructed for the 3rd rows
        # and is rank 1 by construction. In practice, thus far, for each c and
        # (i,j) we either have Qij^c or -Qij^c independently.
        c_mat = np.zeros((3, self.n_img, 3, self.n_img, 3), dtype=self.dtype)
        for i in range(self.n_img - 1):
            for j in range(i + 1, self.n_img):
                ij = self.pairs_to_linear[i, j]
                c_mat[c_vec[ij], i, :, j, :] = rr[ij]

        c_mat = c_mat + c_mat.transpose(0, 3, 4, 1, 2)

        for c in range(3):
            for i in range(self.n_img):
                c_mat[c, i, :, i, :] = c_mat_5d[i, i, c]

        return c_mat, c_mat_5d, c_mat_4d

    def _compute_signs(self, c_mat_5d, c_mat_4d):
        """
        Compute signs for adjusting `c_mat` to be composed of all rank-1 3x3 blocks.
        """
        # To decompose cMat as a rank 1 matrix we need to adjust the signs of the
        # Qij^c so that sign(Qij^c*Qjk^c) = sign(Qik^c) for all c=1,2,3 and (i,j).
        # In practice we compare the sign of the sum of the entries of Qij^c*Qjk^c
        # to the sum of entries of Qik^c.

        # For computational comfort the signs for each c=1,2,3 are stored in a
        # Nx(N over 2) array, where the ij'th column corresponds to the signs of
        # Qij^c * Qjk^c for k~=i,j. The entries in the k=i,j rows of the ij'th
        # column are zero, the value zero is arbitrary, since these entries are
        # not used by the algorithm, and only exist for comfort (of storage and
        # access).
        signs = np.zeros((3, self.n_pairs, self.n_img), dtype=self.dtype)
        for c in range(3):
            for p in range(self.n_pairs):
                i, j = self.pairs[p]
                idx_mask = np.full(self.n_img, True)
                idx_mask[[i, j]] = False
                signs[c, p, idx_mask] = self._calc_Rij_prods(c_mat_5d, i, j, c)

        # Now compute the signs of Qij^c.
        est_signs = np.sign(np.sum(c_mat_4d, axis=(-2, -1)))
        signs = np.transpose(signs, (0, 2, 1))
        for c in range(3):
            signs[c] = est_signs[:, c] * signs[c]

        # Qik^c can be compared with Qir^c*Qrk^c for each r~=i,k, that is,
        # N-2 options. Another way to look at this, is that the r'th image
        # participates in all comparisons of the form sign(Qir^c*Qrk^c)~sign(Qik)
        # for r~=i,k for each c=1,2,3 (see Section 8 in D2 paper).
        # For each image r construct a 3Nx3N matrix. If
        # sign(Qir^c*Qrk^c)~sign(Qik)=1, its ik'th 3x3 block is set to Qik,
        # otherwise, it is set to -Qik.
        sync_signs2 = np.arange(self.n_img).reshape((1, 1, self.n_img, 1))
        sync_signs2 = np.tile(sync_signs2, (3, self.n_img, 1, self.n_img))
        for c in range(3):
            for r in range(self.n_img):
                # Fill signs for synchroniztion for the r'th image.
                # Go over all i,j~=r.
                i_idx = np.concatenate(
                    (np.arange(0, r), np.arange(r + 1, self.n_img))
                )  # i~=r
                for i in i_idx:
                    if i <= r:
                        j_idx = np.concatenate(
                            (np.arange(i + 1, r), np.arange(r + 1, self.n_img))
                        )
                    else:
                        j_idx = np.arange(i + 1, self.n_img)
                    for j in j_idx:
                        ij = self.pairs_to_linear[i, j]
                        sync_signs2[c, r, j, i] = (
                            j + 0.5 * (1 - signs[c, r, ij]) * self.n_img
                        )
                        sync_signs2[c, r, i, j] = (
                            i + 0.5 * (1 - signs[c, r, ij]) * self.n_img
                        )
                        # The function (1-x)/2 maps 1->0 and -1->1

        return sync_signs2

    def _estimate_rows(self, sync_signs2, c_mat_5d):
        """
        Construct 3N x 3N matrix of rank-1 3x3 blocks of sij*vi_m.T @ vj_m,
        the leading eigenvectors of which correspond to estimates for the rows
        of the rotations Ri, up to signs.
        """
        c_mat_5d_mp = np.concatenate((c_mat_5d, -c_mat_5d), axis=1)
        rows_arr = np.zeros((3, self.n_img, 3 * self.n_img), dtype=self.dtype)
        svals = np.zeros((3, 2, self.n_img), dtype=self.dtype)

        logger.info("Constructing and decomposing N sign synchronization matrices...")
        for c in range(3):
            for r in range(self.n_img):
                # Image r used for signs.
                c_mat_eff = self._fill_sign_sync_matrix_c(
                    c_mat_5d_mp, sync_signs2, c, r
                )

                # Construct (3*N)x(3*N) rank 1 matrices from Qik
                c_mat_for_svd = np.zeros(
                    (3 * self.n_img, 3 * self.n_img), dtype=self.dtype
                )
                for i in range(self.n_img):
                    row_3Nx3 = c_mat_eff[i]
                    row_3Nx3 = row_3Nx3.reshape(3 * self.n_img, 3)
                    c_mat_for_svd[:, 3 * i : 3 * i + 3] = row_3Nx3

                c_mat_for_svd = c_mat_for_svd + c_mat_for_svd.T

                # Extract leading eigenvector of rank 1 matrix. For each r and c
                # this gives an estimate for the c'th row of the rotation Rr, up
                # to sign +/-.
                for i in range(self.n_img):
                    c_mat_for_svd[3 * i : 3 * i + 3, 3 * i : 3 * i + 3] = c_mat_eff[
                        i, i
                    ]
                U, S, _ = np.linalg.svd(c_mat_for_svd)
                svals[c, :, r] = S[:2]
                rows_arr[c, r] = U[:, 0]

        return rows_arr

    def _compute_signs_adjustment(self, rows_arr):
        """
        Compute signs adjustment vector.
        """
        # Sync signs according to results for each image. Dot products between
        # signed row estimates are used to construct an (N over 2)x(N over 2)
        # sign synchronization matrix S. If (v_i)k and (v_j)k are the i'th and
        # j'th estimates for the c'th row of Rk, then the entry (i,k),(k,j) entry
        # of S is <(v_i)k,(v_j)k>, where the rows and columns of S are indexed by
        # double indexes (i,j), 1<=i<j<=(N over 2).
        pairs_map = np.zeros((self.n_pairs, 2 * (self.n_img - 2)), dtype=int)
        for i in range(self.n_img):
            for j in range(i + 1, self.n_img):
                ij = self.pairs_to_linear[i, j]
                pairs_map[ij] = np.concatenate(
                    (
                        self.pairs_to_linear[:i, i],
                        self.pairs_to_linear[i, np.r_[i + 1 : j, j + 1 : self.n_img]],
                        self.pairs_to_linear[np.r_[:i, i + 1 : j], j],
                        self.pairs_to_linear[j, j + 1 :],
                    )
                )

        signs = np.zeros((3, self.n_pairs), dtype=self.dtype)
        s_out = np.zeros((3, 3), dtype=self.dtype)

        logger.info("Constructing and decomposing 3 sign synchroniztion matrices.")
        # The matrix S requires space on order of O(N^4). Instead of storing it
        # in memory we compute its SVD using the function smat which multiplies
        # (N over 2)x1 vectors by S.
        for c in range(3):
            # Prepare data for smat to act on vectors.
            sign_mat = np.zeros((self.n_pairs, 2 * (self.n_img - 2)), dtype=int)
            for i in range(self.n_img - 1):
                for j in range(i + 1, self.n_img):
                    ij = self.pairs_to_linear[i, j]
                    sij = rows_arr[c, j, 3 * i : 3 * i + 3]
                    sji = rows_arr[c, i, 3 * j : 3 * j + 3]
                    siks = rows_arr[
                        c, np.r_[:i, i + 1 : j, j + 1 : self.n_img], 3 * i : 3 * i + 3
                    ]
                    sjks = rows_arr[
                        c, np.r_[:i, i + 1 : j, j + 1 : self.n_img], 3 * j : 3 * j + 3
                    ]
                    sign_mat[ij] = np.concatenate(
                        (np.sign(siks @ sij), np.sign(sjks @ sji))
                    )

            smat = la.LinearOperator(
                shape=(self.n_pairs, self.n_pairs),
                matvec=lambda v, s=sign_mat: self._mult_smat_by_vec(v, s, pairs_map),
                rmatvec=lambda v, s=sign_mat: self._mult_smat_by_vec(v, s, pairs_map),
            )
            U, S, _ = la.svds(smat, k=3, which="LM")
            U = np.sign(U[0]) * U  # Stable svds
            signs[c] = U[:, -1]  # svds returns in ascending order
            s_out[c] = S[::-1]

        return np.sign(signs)

    def _extract_rotations(self, c_mat, signs):
        """
        Adjust the signs of each block of `c_mat` then extract the rotation
        rows and construct the estimated rotations.

        :param c_mat: The color synchronization matrix.
        :param signs: The signs adjustment matrix.
        :return: Estimated rotations.
        """
        # Adjust the signs of Qij^c in the matrices cMat(:,:,c) for all c=1,2,3
        # and 1<=i<j<=N according to the results of the signs from the last stage.
        logger.info("Constructing and decomposing 3 row synchroniztion matrices.")
        for c in range(3):
            idx = 0
            for i in range(self.n_img - 1):
                for j in range(i + 1, self.n_img):
                    c_mat[c, j, :, i, :] *= signs[c, idx]
                    c_mat[c, i, :, j, :] *= signs[c, idx]
                    idx += 1

        # cMat(:,:,c) are now rank 1. Decompose using SVD and take leading eigenvector.
        c_mat = c_mat.reshape(3, 3 * self.n_img, 3 * self.n_img)
        U1, _, _ = la.svds(c_mat[0], k=3, which="LM")
        U2, _, _ = la.svds(c_mat[1], k=3, which="LM")
        U3, _, _ = la.svds(c_mat[2], k=3, which="LM")

        # Stabilize and take leading eigenvector.
        U1 = np.sign(U1[0, -1]) * U1[:, -1]
        U2 = np.sign(U2[0, -1]) * U2[:, -1]
        U3 = np.sign(U3[0, -1]) * U3[:, -1]

        # The c'th row of the rotation Rj is Uc(3*j-2:3*j,1)/norm(Uc(3*j-2:3*j,1)),
        # (Rows must be normalized to length 1).
        logger.info("Assembeling rows to rotations matrices.")
        rot = np.zeros((self.n_img, 3, 3), dtype=self.dtype)
        rot[:, 0] = U1.reshape(self.n_img, 3)
        rot[:, 1] = U2.reshape(self.n_img, 3)
        rot[:, 2] = U3.reshape(self.n_img, 3)
        rot /= np.linalg.norm(rot, axis=-1)[:, :, None]

        # Ensure we have rotations.
        not_a_rot = np.argwhere(np.linalg.det(rot) < 0)
        rot[not_a_rot, 2] *= -1

        return rot

    def _fill_sign_sync_matrix_c(self, c_mat_5d_mp, sync_signs2, c, img):
        c_mat_eff = np.zeros((self.n_img, self.n_img, 3, 3), dtype=self.dtype)
        for r in range(self.n_img):
            c_mat_eff[:, r] = c_mat_5d_mp[r, sync_signs2[c, img, :, r], c]
        return c_mat_eff

    def _calc_Rij_prods(self, c_mat_5d, i, j, c):
        Rik = np.delete(c_mat_5d[i, :, c], [i, j], axis=0)
        Rkj = np.delete(c_mat_5d[:, j, c], [i, j], axis=0)
        Rij = Rik @ Rkj

        # In case we get a zero score arbitrarily choose sign +1.
        ij_signs = np.sum(Rij, axis=(-2, -1))
        zeros_idx = ij_signs == 0
        if np.count_nonzero(zeros_idx) > 0:
            ij_signs[zeros_idx] = 1

        return np.sign(ij_signs)

    def _mult_smat_by_vec(self, v, sign_mat, pairs_map):
        """
        Multiplies the signs sync matrix by a vector.
        """
        v_out = np.zeros_like(v)
        for i in range(self.n_img):
            for j in range(i + 1, self.n_img):
                ij = self.pairs_to_linear[i, j]
                v_out[ij] = sign_mat[ij] @ v[pairs_map[ij]]
        return v_out

    ####################
    # Helper Functions #
    ####################

    @staticmethod
    def _circ_seq(n1, n2, L):
        """
        For integers 0 <= n1, n2 < L, make a circular sequence of integers between
        n1 and n2 modulo L.

        :param n1: First integer in sequence.
        :param n2: Last integer in sequence.
        :param L: Modulus of values in sequence.
        :return: Circular sequence modulo L.
        """
        if min(n1, n2) < 0 or max(n1, n2) >= L:
            raise ValueError(
                f"n1 and n2 must both be in [0, {L}). Found n1={n1}, n2={n2}."
            )
        if n2 < n1:
            n2 += L
        if n1 == n2:
            return np.array([n1]).astype(int) % L

        seq = np.arange(n1, n2 + 1).astype(int) % L

        return seq

    @staticmethod
    def _saff_kuijlaars(N):
        """
        Generates N vertices on the unit sphere that are approximately evenly distributed.

        This implements the recommended algorithm in spherical coordinates
        (theta, phi) according to "Distributing many points on a sphere"
        by E.B. Saff and A.B.J. Kuijlaars, Mathematical Intelligencer 19.1
        (1997) 5--11.

        :param N: Number of vertices to generate.

        :return: Nx3 array of vertices in cartesian coordinates.
        """
        k = np.arange(1, N + 1)
        h = -1 + 2 * (k - 1) / (N - 1)
        theta = np.arccos(h)
        phi = np.zeros(N)

        for i in range(1, N - 1):
            phi[i] = (phi[i - 1] + 3.6 / (np.sqrt(N * (1 - h[i] ** 2)))) % (2 * np.pi)

        # Spherical coordinates
        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)

        mesh = np.column_stack((x, y, z))

        return mesh

    @staticmethod
    def _mark_equators(sphere_grid, eq_filter_angle):
        """
        This method categorizes a set of 3D unit vectors into equator and non-equator
        vectors determined by the parameter `eq_filter_angle`, returned as `eq_idx`.
        It further categorizes the vectors into the classes non_equator, z-equator,
        y-equator, x-equator, z-top_view, y-top_view, and x-top_view, which are labeled
        respectively with the values 0 - 6 and returned as `eq_class`.

        :param sphere_grid: Nx3 array of vertices in cartesian coordinates.
        :param eq_filter_angle: Angular distance from equator to be marked as
            an equator point.

        :return: eq_class, n_rots length array of values indicating equator class.
        """
        # Project each vector onto xy, xz, yz planes and measure angular distance
        # from each plane.
        n_rots = len(sphere_grid)
        angular_dists = np.zeros((n_rots, 3), dtype=sphere_grid.dtype)

        # For each grid point get the distance from the z, y, and x-axis equators.
        for i in range(3):
            proj_along_axis = sphere_grid.copy()
            proj_along_axis[:, 2 - i] = 0
            proj_along_axis /= np.linalg.norm(proj_along_axis, axis=1)[:, None]
            angular_dists[:, i] = np.sum(sphere_grid * proj_along_axis, axis=-1)

        # Mark all views close to an equator.
        eq_min_dist = np.cos(eq_filter_angle * np.pi / 180)
        n_eqs = np.count_nonzero(angular_dists > eq_min_dist, axis=1)

        # Classify equators.
        # 0 -> non-equator view
        # 1 -> z equator
        # 2 -> y equator
        # 3 -> x equator
        # 4 -> z top view
        # 5 -> y top view
        # 6 -> x top view
        eq_class = np.zeros(n_rots)

        # Grid points which are equator points with respect to 2 equators are considered top views.
        # For example, a grid point that is close to both the x and y equator is a z top view.
        top_view_idx = n_eqs > 1
        top_view_class = np.argmin(angular_dists[top_view_idx] > eq_min_dist, axis=1)
        eq_class[top_view_idx] = top_view_class + 4

        # Assign grid points which are equator points with respect to only 1 equator.
        eq_view_idx = n_eqs == 1
        eq_view_class = np.argmax(angular_dists[eq_view_idx] > eq_min_dist, axis=1)
        eq_class[eq_view_idx] = eq_view_class + 1

        return eq_class

    @staticmethod
    def _generate_inplane_rots(sphere_grid, d_theta):
        """
        This function takes projection directions (points on the 2-sphere) and
        generates rotation matrices in SO(3). The projection direction
        is the 3rd column and columns 1 and 2 span the perpendicular plane.
        To properly discretize SO(3), for each projection direction we generate
        [2*pi/dtheta] "in-plane" rotations, of the plane
        perpendicular to this direction. This is done by generating one rotation
        for each direction and then multiplying on the right by a rotation about
        the Z-axis by k*dtheta degrees, k=0...2*pi/dtheta-1.

        :param sphere_grid: A set of points on the 2-sphere.
        :param d_theta: Resolution for in-plane rotations (in degrees)
        :returns: 4D array of rotations of size len(sphere_grid) x n_inplane_rots x 3 x 3.
        """
        dtype = sphere_grid.dtype
        # Generate one rotation for each point on the sphere.
        n_rots = len(sphere_grid)
        Ri2 = np.column_stack((-sphere_grid[:, 1], sphere_grid[:, 0], np.zeros(n_rots)))
        Ri2 /= np.linalg.norm(Ri2, axis=1)[:, None]
        Ri1 = np.cross(Ri2, sphere_grid)
        Ri1 /= np.linalg.norm(Ri1, axis=1)[:, None]

        rots_grid = np.zeros((n_rots, 3, 3), dtype=dtype)
        rots_grid[:, :, 0] = Ri1
        rots_grid[:, :, 1] = Ri2
        rots_grid[:, :, 2] = sphere_grid

        # Generate in-plane rotations.
        d_theta *= np.pi / 180
        # Negative signs to match matlab.
        inplane_rots = Rotation.about_axis(
            "z", np.arange(0, -2 * np.pi, -d_theta), dtype=dtype
        ).matrices
        n_inplane_rots = len(inplane_rots)

        # Generate in-plane rotations of rots_grid.
        inplane_rotated_grid = np.zeros((n_rots, n_inplane_rots, 3, 3), dtype=dtype)
        for i in range(n_rots):
            inplane_rotated_grid[i] = rots_grid[i] @ inplane_rots

        return inplane_rotated_grid

    def _generate_commonline_indices(self, cl_angles):
        """
        Converts a multi-dimensional stack of pairs of commonline angles in [0, 360) degrees
        into a flattened stack of polar Fourier linear indices, with the convention that
        each linear index corresponds to an unraveled index in [0, n_theta // 2) x [0, n_theta).

        :param cl_angles: A multi-dimensional stack of commonline angles in degrees, shape (..., 2).
        :return: cl_idx, a 1D array of linear indices.
        """
        L = self.n_theta

        # Flatten the stack
        og_shape = cl_angles.shape
        cl_angles = np.reshape(cl_angles, (np.prod(og_shape[:-1]), 2))

        # Fourier ray index
        row_sub = np.round(cl_angles[:, 0] * L / 360).astype("int") % L
        col_sub = np.round(cl_angles[:, 1] * L / 360).astype("int") % L

        # Restrict Ri in-plane coordinates to <180 degrees.
        is_geq_than_pi = row_sub >= L // 2
        row_sub[is_geq_than_pi] = row_sub[is_geq_than_pi] - L // 2
        col_sub[is_geq_than_pi] = (col_sub[is_geq_than_pi] + (L // 2)) % L

        # Convert to linear indices in 180x360 correlation matrix.
        cl_idx = np.ravel_multi_index((row_sub, col_sub), dims=(L // 2, L))

        return cl_idx



import logging

import numpy as np
from scipy.sparse.linalg import LinearOperator

from aspire.image import Image
from aspire.numeric.scipy import cg
from aspire.utils import mdim_mat_fun_conj
from aspire.volume import Volume

logger = logging.getLogger(__name__)


class Coef:
    """
    Numpy interoperable container for stacks of real coefficient vectors.
    Each `Coef` instance has an associated `Basis`.
    """

    _allowed_dtypes = (np.float32, np.float64)

    def __init__(self, basis, data, dtype=None):
        """
        A stack of one or more coefficient arrays.

        The stack can be multidimensional with `stack_size` equal
        to the product of the stack dimensions.  Singletons will be
        expanded into a 1D stack of length one.

        The last axes always represents the coefficient `count`.

        :param basis: `Basis` associated with `data` coefficients.
        :param data: Numpy array containing image data with shape
            `(..., count)`.
        :param dtype: Optionally cast `data` to this dtype.
            Defaults to `data.dtype`.

        :return: `Coef` instance holding `data`.
        """

        if not isinstance(data, np.ndarray):
            raise ValueError("Coef should be instantiated with an ndarray")

        if data.ndim < 1:
            raise ValueError(
                "Coef data should be ndarray with shape (N1...) x count or (count)."
            )
        elif data.ndim == 1:
            data = np.expand_dims(data, axis=0)

        if dtype is None:
            self.dtype = data.dtype
        else:
            self.dtype = np.dtype(dtype)

        # Check real/complex dtype based on class.
        self._check_dtype()

        if not isinstance(basis, Basis):
            raise TypeError(
                f"`basis` is required to be a `Basis` instance, received {type(basis)}"
            )
        self.basis = basis

        self._data = data.astype(self.dtype, copy=False)
        self.ndim = self._data.ndim
        self.shape = self._data.shape
        self.stack_ndim = self._data.ndim - 1
        self.stack_shape = self._data.shape[:-1]
        self.stack_size = np.prod(self.stack_shape)
        self.count = self._data.shape[-1]

        # Derive count from basis.
        basis_count = self._get_basis_count()

        if self.count != basis_count:
            raise RuntimeError(
                f"Provided data count of {self.count} does not match basis count of {basis_count}."
            )

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self.asnumpy().__array_interface__
        self.__array__ = self.asnumpy()

    def _check_dtype(self):
        """
        Private helper method to check real/complex dtype based on class `_allowed_dtypes`.

        Raises on mismatch.
        """

        if self.dtype not in self._allowed_dtypes:
            raise TypeError(
                f"{self.__class__.__name__} requires {self._allowed_dtypes} coefficients, attempted {self.dtype}."
            )

    def _get_basis_count(self):
        """
        Private helper method to return coefficient count from basis.

        :return: Basis count (integer).
        """
        return int(self.basis.count)

    def __len__(self):
        """
        Return length of slowest stack axis.
        """
        return self.stack_shape[0]

    def asnumpy(self):
        """
        Return image data as a (<stack>, count)
        read-only array view.

        :return: read-only ndarray view
        """

        view = self._data.view()
        view.flags.writeable = False
        return view

    def _check_key_dims(self, key):
        if isinstance(key, tuple) and (len(key) > self._data.ndim):
            raise ValueError(
                f"Coef stack_dim is {self.stack_ndim}, slice length must be =< {self.ndim}"
            )

    def __getitem__(self, key):
        self._check_key_dims(key)
        return self.__class__(self.basis, self._data[key])

    def __setitem__(self, key, value):
        self._check_key_dims(key)
        self._data[key] = value

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :returns: Coef instance
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.stack_size:
            raise ValueError(
                f"Number of coefficient vectors {self.stack_size} cannot be reshaped to {shape}."
            )

        return self.__class__(
            self.basis, self._data.reshape(*shape, self._data.shape[-1])
        )

    def copy(self):
        """
        Return a new `Coef` instance with a deep copy of the data.
        """
        return self.__class__(self.basis, self._data.copy())

    def evaluate(self):
        """
        Return the evaluation of coefficients in the associated `basis`.
        """
        return self.basis.evaluate(self)

    def rotate(self, radians, refl=None):
        """
        Returns coefs rotated counter-clockwise by `radians`.

        Raises error if underlying coef basis does not support rotations.

        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: rotated coefs.
        """

        if not callable(getattr(self.basis, "rotate", None)):
            raise RuntimeError(
                f"self.basis={self.basis} does not provide `rotate` method."
            )

        return self.basis.rotate(self, radians, refl)

    def shift(self, shifts):
        """
        Returns coefs shifted by `shifts`.

        This will transform to real cartesian space, shift,
        and transform back to basis space.

        :param coef: Basis coefs.
        :param shifts: Shifts in pixels (x,y). Shape (1,2) or (len(coef), 2).
        :return: coefs of shifted images.
        """

        if not callable(getattr(self.basis, "shift", None)):
            raise RuntimeError(
                f"self.basis={self.basis} does not provide `shift` method."
            )

        return self.basis.shift(self, shifts)

    def __mul__(self, other):
        """
        Overload operator for multiplication.

        :param other: `Coef` instance to multiply with.
            Also allows for multiplication by Numpy arrays and scalars.
        :return: `Coef` instance.
        """

        if isinstance(other, Coef):
            other = other._data

        return self.__class__(self.basis, self._data * other)

    def __add__(self, other):
        """
        Overload operator for addition.

        :param other: `Coef` instance to add.
            Also allows for addition by Numpy arrays and scalars.
        :return: `Coef` instance.
        """

        if isinstance(other, Coef):
            other = other._data

        return self.__class__(self.basis, self._data + other)

    def __sub__(self, other):
        """
        Overload operator for subtraction.

        :param other: `Coef` instance to subtract.
            Also allows for subtraction by Numpy arrays and scalars.
        :return: `Coef` instance.
        """

        if isinstance(other, Coef):
            other = other._data

        return self.__class__(self.basis, self._data - other)

    def __neg__(self):
        """
        Overload operator for negation.

        :return: `Coef` instance.
        """

        return self.__class__(self.basis, -self._data)

    @property
    def size(self):
        """
        Return np.size of underlying data.

        This should be `stack_size * count`,
        or `len(self) * count`.
        """
        return np.size(self._data)

    # This is included for completion, but is not being adopted yet.
    def by_indices(self, **kwargs):
        """
        Select coefficients by indices (`radial`, `angular`).

        See `SteerableBasis.indices_mask` for argument details.

        :return: Numpy array.
        """

        mask = self.basis.indices_mask(**kwargs)
        return self._data[..., mask]

    def to_complex(self):
        """
        Convert and return real coefficients as `ComplexCoef`.
        """
        return self.basis.to_complex(self)

    def to_real(self):
        """
        Not implemented for real Coef.
        """
        raise TypeError("Coef already real.")


class ComplexCoef(Coef):
    """
    Numpy interoperable container for stacks of complex coefficient vectors.
    Each `ComplexCoef` instance has an associated `Basis`.
    """

    _allowed_dtypes = (np.complex64, np.complex128)

    def _get_basis_count(self):
        """
        Private helper method to return coefficient complex count from basis.

        :return: Basis complex count (integer).
        """

        return int(self.basis.complex_count)

    def evaluate(self):
        """
        Return the evaluation of coefficients in the associated `basis`.
        """
        return self.to_real().evaluate()

    def rotate(self, radians, refl=None):
        """
        Returns coefs rotated counter-clockwise by `radians`.

        Raises error if underlying coef basis does not support rotations.

        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: Rotated ComplexCoefs.
        """

        return self.to_real().rotate(radians, refl).to_complex()

    def shift(self, shifts):
        """
        Returns complex coefs shifted by `shifts`.

        This will transform to real cartesian space, shift,
        and transform back to basis space.

        :param coef: Basis coefs.
        :param shifts: Shifts in pixels (x,y). Shape (1,2) or (len(coef), 2).
        :return: Complex coefs of shifted images.
        """

        return self.to_real().shift(shifts).to_complex()

    def to_real(self):
        """
        Convert and return complex coefficients as `Coef`.
        """
        return self.basis.to_real(self)

    def to_complex(self):
        """
        Not implemented for ComplexCoef.
        """
        raise TypeError("ComplexCoef already complex.")


class Basis:
    """
    Define a base class for expanding 2D particle images and 3D structure volumes

    """

    def __init__(self, size, ell_max=None, dtype=np.float32):
        """
        Initialize an object for the base of basis class

        :param size: The size of the vectors for which to define the basis.
            Currently only square images and cubic volumes are supported.
        :param ell_max: The maximum order ell of the basis elements. If no input
            (= None), it will be set to np.Inf and the basis includes all
            ell such that the resulting basis vectors are concentrated
            below the Nyquist frequency (default Inf).
        """
        if ell_max is None:
            ell_max = np.inf

        ndim = len(size)
        nres = size[0]
        self.sz = size
        self.nres = nres
        self.count = 0
        self.ell_max = ell_max
        self.ndim = ndim
        if self.ndim == 2:
            self._cls = Image
        elif self.ndim == 3:
            self._cls = Volume
        else:
            raise RuntimeError("Basis ndim must be 2 or 3")
        self.dtype = np.dtype(dtype)
        if self.dtype not in (np.float32, np.float64):
            raise NotImplementedError(
                "Currently only implemented for float32 and float64 types"
            )
        # dtype of coefficients is the same as self.dtype for real bases
        # subclasses with complex coefficients override this attribute
        self.coefficient_dtype = self.dtype

        self._build()

    def _build(self):
        """
        Build the internal data structure to represent basis
        """
        raise NotImplementedError("subclasses must implement this")

    def _precomp(self):
        """
        Precompute the basis functions at defined sample points
        """
        raise NotImplementedError("subclasses must implement this")

    def norms(self):
        """
        Calculate the normalized factors of basis functions
        """
        raise NotImplementedError("subclasses must implement this")

    def evaluate(self, v):
        """
        Evaluate coefficient vector in basis

        :param v: `Coef` instance containing the coefficients to be
            evaluated. The first dimension must correspond to the
            number of coefficient vectors, while the second must
            correspond to `self.count`.
        :return: The evaluation of the coefficient vector(s) `v` for this basis.
            This is an Image or a Volume object containing one image/volume for each
            coefficient vector, and of size `self.sz`.
        """

        if v.dtype != self.coefficient_dtype:
            logger.warning(
                f"{self.__class__.__name__}::evaluate"
                f" Inconsistent dtypes v: {v.dtype} self coefficient dtype: {self.coefficient_dtype}"
            )

        if not isinstance(v, Coef):
            raise TypeError(f"`evaluate` should be passed a `Coef`, received {type(v)}")

        # Flatten stack
        stack_shape = v.stack_shape
        v = v.stack_reshape(-1).asnumpy()

        # Compute the transform
        x = self._evaluate(v)
        # Restore stack shape
        x = x.reshape(*stack_shape, *self.sz)

        # Return the appropriate class
        return self._cls(x)

    def _evaluate(self, v):
        raise NotImplementedError("subclasses must implement this")

    def evaluate_t(self, v):
        """
        Evaluate coefficient in dual basis

        :param v: An Image or Volume object whose size matches `self.sz`.
        :return: The evaluation of the Image or Volume object `v` in the dual
            basis of `basis`.
            This is an array of vectors whose first dimension equals the number of
            images/volumes in `v`. and whose second dimension is `self.count`.
        """
        if v.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}::evaluate_t"
                f" Inconsistent dtypes v: {v.dtype} self: {self.dtype}"
            )

        if not isinstance(v, self._cls):
            logger.warning(
                f"{self.__class__.__name__}::evaluate_t"
                f" passed numpy array instead of {self._cls}."
            )
        else:
            v = v.asnumpy()

        # Flatten stack, ndim is wrt Basis (2 or 3)
        stack_shape = v.shape[: -self.ndim]
        v = v.reshape(-1, *v.shape[-self.ndim :])
        # Compute the adjoint
        x = self._evaluate_t(v)
        # Restore stack shape
        x = x.reshape(*stack_shape, self.count)

        return Coef(self, x)

    def _evaluate_t(self, v):
        raise NotImplementedError("Subclasses should implement this")

    def mat_evaluate(self, V):
        """
        Evaluate coefficient matrix in basis

        :param V: A coefficient matrix of size `self.count`-by-
            `self.count` to be evaluated.
        :return: A multidimensional matrix of size `self.sz`-by
            -`self.sz` corresponding to the evaluation of `V` in
            this basis.
        """
        return mdim_mat_fun_conj(V, 1, len(self.sz), self._evaluate)

    def mat_evaluate_t(self, X):
        """
        Evaluate coefficient matrix in dual basis

        :param X: The coefficient array of size `self.sz`-by-`self.sz`
            to be evaluated.
        :return: The evaluation of `X` in the dual basis. This is
            `self.count`-by-`self.count`. matrix.
            If `V` is a matrix of size `self.count`-by-`self.count`,
            `B` is the change-of-basis matrix of `basis`, and `x` is a
            multidimensional matrix of size `basis.sz`-by-`basis.sz`, the
            function calculates V = B' * X * B, where the rows of `B`, rows
            of 'X', and columns of `X` are read as vectorized arrays.
        """
        return mdim_mat_fun_conj(X, len(self.sz), 1, self._evaluate_t)

    def expand(self, x, tol=None, atol=0):
        """
        Obtain coefficients in the basis from those in standard coordinate basis

        This is a similar function to evaluate_t but with more accuracy by using
        the cg optimizing of linear equation, Ax=b.

        :param x: An array whose last two or three dimensions are to be expanded
            the desired basis. These dimensions must equal `self.sz`.
        :param tol: Relative tolerance for convergence, `norm(residual) <= max(tol*norm(b), atol)`.
            Deafult `None` sets to dtype's `eps`*10.
        :param atol: Absolute tolerance for convergence, `norm(residual) <= max(tol*norm(b), atol)`.
        :return: The coefficients of `v` expanded in the desired basis.
            The last dimension of `v` is with size of `count` and the
            first dimensions of the return value correspond to
            those first dimensions of `x`.

        """

        if isinstance(x, Image) or isinstance(x, Volume):
            x = x.asnumpy()

        if x.dtype != self.dtype:
            logger.warning(
                f"{self.__class__.__name__}::expand"
                f" Inconsistent dtypes x: {x.dtype} self: {self.dtype}"
            )

        # TODO: We should  only need to do this block when we are not passed Image/Volume.
        # check that last ndim values of input shape match
        # the shape of this basis
        assert (
            x.shape[-self.ndim :] == self.sz
        ), f"Last {self.ndim} dimensions of x must match {self.sz}."
        # extract number of images/volumes, or () if only one
        sz_roll = x.shape[: -self.ndim]
        # convert to standardized shape e.g. (L,L) to (1,L,L)
        x = x.reshape((-1, *self.sz))

        operator = LinearOperator(
            shape=(self.count, self.count),
            matvec=lambda v: self.evaluate_t(self.evaluate(Coef(self, v))),
            dtype=self.dtype,
        )

        if tol is None:
            # TODO: (from MATLAB implementation) - Check that this tolerance make sense for multiple columns in v
            tol = 10 * np.finfo(x.dtype).eps
        logger.info(f"Expanding array in basis with tol={tol} atol={atol}")

        # number of image samples
        n_data = x.shape[0]
        v = np.zeros((n_data, self.count), dtype=self.coefficient_dtype)

        for isample in range(0, n_data):
            b = self.evaluate_t(self._cls(x[isample])).asnumpy().T
            # TODO: need check the initial condition x0 can improve the results or not.
            v[isample], info = cg(operator, b, rtol=tol, atol=atol)
            if info != 0:
                raise RuntimeError(f"Unable to converge! cg info={info}")

        # return v coefficients with the last dimension of self.count
        v = v.reshape((*sz_roll, self.count))

        return Coef(self, v)



"""
General purpose math functions, mostly geometric in nature.
"""

import logging
import math

import numpy as np
from numpy.linalg import norm
from scipy.linalg import svd

from aspire.numeric import xp
from aspire.utils.random import Random
from aspire.utils.rotation import Rotation

logger = logging.getLogger(__name__)


def cart2pol(x, y):
    """
    Convert Cartesian to Polar Coordinates. All input arguments must be the same shape.

    :param x: x-coordinate in Cartesian space
    :param y: y-coordinate in Cartesian space
    :return: A 2-tuple of values:
        theta: angular coordinate/azimuth
        r: radial distance from origin
    """
    return np.arctan2(y, x), np.hypot(x, y)


def cart2sph(x, y, z):
    """
    Transform cartesian coordinates to spherical. All input arguments must be the same shape.

    :param x: X-values of input co-ordinates.
    :param y: Y-values of input co-ordinates.
    :param z: Z-values of input co-ordinates.
    :return: A 3-tuple of values, all of the same shape as the inputs.
        (<azimuth>, <elevation>, <radius>)
        azimuth and elevation are returned in radians.

    This function is equivalent to MATLAB's cart2sph function.
    """
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r


def _mgrid_slice(n, shifted, normalized):
    """
    Util to generate a `slice` representing a 1d linspace
    as expected by `np.mgrid`.

    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :return: `slice` to be used by `np.mgrid`.
    """

    num_points = n * 1j
    start = -n // 2 + 1
    end = n // 2

    if shifted and n % 2 == 0:
        start -= 1 / 2
        end -= 1 / 2
    elif n % 2 == 0:
        start -= 1
        end -= 1

    if normalized:
        # Compute the denominator for normalization
        denom = n / 2
        if shifted and n % 2 == 0:
            denom -= 1 / 2

        # Apply the normalization
        start /= denom
        end /= denom

    return slice(start, end, num_points)


def grid_1d(n, shifted=False, normalized=True, dtype=np.float32):
    """
    Generate one dimensional grid.

    :param n: the number of grid points.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :return: the rectangular and polar coordinates of all grid points.
    """

    r = x = np.mgrid[_mgrid_slice(n, shifted, normalized)].astype(dtype)

    return {"x": x, "r": r}


def grid_2d(n, shifted=False, normalized=True, indexing="yx", dtype=np.float32):
    """
    Generate two dimensional grid.

    :param n: the number of grid points in each dimension.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :param indexing: 'yx' (C) or 'xy' (F), defaulting to 'yx'.
        See https://numpy.org/doc/stable/reference/generated/numpy.meshgrid.html
    :return: the rectangular and polar coordinates of all grid points.
    """

    grid = _mgrid_slice(n, shifted, normalized)
    y, x = np.mgrid[grid, grid].astype(dtype)
    if indexing == "xy":
        x, y = y, x
    elif indexing != "yx":
        raise RuntimeError(
            f"grid_2d indexing {indexing} not supported." "  Try 'xy' or 'yx'"
        )

    phi, r = cart2pol(x, y)

    return {"x": x, "y": y, "phi": phi, "r": r}


def grid_3d(n, shifted=False, normalized=True, indexing="zyx", dtype=np.float32):
    """
    Generate three dimensional grid.

    :param n: the number of grid points in each dimension.
    :param shifted: shifted by half of grid or not when n is even.
    :param normalized: normalize the grid in the range of (-1, 1) or not.
    :param indexing: 'zyx' (C) or 'xyz' (F), defaulting to 'zyx'.
        See https://numpy.org/doc/stable/reference/generated/numpy.meshgrid.html
    :return: the rectangular and spherical coordinates of all grid points.
    """

    grid = _mgrid_slice(n, shifted, normalized)
    z, y, x = np.mgrid[grid, grid, grid].astype(dtype)

    if indexing == "xyz":
        x, y, z = z, y, x
    elif indexing != "zyx":
        raise RuntimeError(
            f"grid_3d indexing {indexing} not supported." "  Try 'xyz' or 'zyx'"
        )

    phi, theta, r = cart2sph(x, y, z)

    # TODO: Should this theta adjustment be moved inside cart2sph?
    theta = np.pi / 2 - theta

    return {"x": x, "y": y, "z": z, "phi": phi, "theta": theta, "r": r}


def uniform_random_angles(n, seed=None, dtype=np.float32):
    """
    Generate random 3D rotation angles

    :param n: The number of rotation angles to generate
    :param seed: Random integer seed to use. If None, the current random state is used.
    :return: A n-by-3 ndarray of rotation angles
    """
    # Generate random rotation angles, in radians
    with Random(seed):
        angles = np.column_stack(
            (
                np.random.random(n) * 2 * np.pi,
                np.arccos(2 * np.random.random(n) - 1),
                np.random.random(n) * 2 * np.pi,
            )
        )
    return angles.astype(dtype)


def register_rotations(rots, rots_ref):
    """
    Register estimated orientations to reference ones.

    Finds the orthogonal transformation that best aligns the estimated rotations
    to the reference rotations.

    :param rots: The rotations to be aligned in the form of a n-by-3-by-3 array.
    :param rots_ref: The reference rotations to which we would like to align in
        the form of a n-by-3-by-3 array.
    :return: o_mat, optimal orthogonal 3x3 matrix to align the two sets;
        flag, flag==1 then J conjugacy is required and 0 is not.
    """

    assert (
        rots.shape == rots_ref.shape
    ), "Two sets of rotations must have same dimensions."
    K = rots.shape[0]

    # Reflection matrix
    J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

    Q1 = np.zeros((3, 3), dtype=rots.dtype)
    Q2 = np.zeros((3, 3), dtype=rots.dtype)

    for k in range(K):
        R = rots[k, :, :]
        Rref = rots_ref[k, :, :]
        Q1 = Q1 + R @ Rref.T
        Q2 = Q2 + (J @ R @ J) @ Rref.T

    # Compute the two possible orthogonal matrices which register the
    # estimated rotations to the true ones.
    Q1 = Q1 / K
    Q2 = Q2 / K

    # We are registering one set of rotations (the estimated ones) to
    # another set of rotations (the true ones). Thus, the transformation
    # matrix between the two sets of rotations should be orthogonal. This
    # matrix is either Q1 if we recover the non-reflected solution, or Q2,
    # if we got the reflected one. In any case, one of them should be
    # orthogonal.

    err1 = norm(Q1 @ Q1.T - np.eye(3, dtype=rots.dtype), ord="fro")
    err2 = norm(Q2 @ Q2.T - np.eye(3, dtype=rots.dtype), ord="fro")

    # In any case, enforce the registering matrix O to be a rotation.
    if err1 < err2:
        # Use Q1 as the registering matrix
        U, _, V = svd(Q1)
        flag = 0
    else:
        # Use Q2 as the registering matrix
        U, _, V = svd(Q2)
        flag = 1

    Q_mat = U @ V

    return Q_mat, flag


def get_aligned_rotations(rots, Q_mat, flag):
    """
    Get aligned rotation matrices to reference ones.

    Calculated aligned rotation matrices from the orthogonal transformation
    that best aligns the estimated rotations to the reference rotations.

    :param rots: The reference rotations to which we would like to align in
        the form of a n-by-3-by-3 array.
    :param Q_mat:  optimal orthogonal 3x3 transformation matrix
    :param flag:  flag==1 then J conjugacy is required and 0 is not
    :return: regrot, aligned rotation matrices
    """

    K = rots.shape[0]

    # Reflection matrix
    J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

    regrot = np.zeros_like(rots)
    for k in range(K):
        R = rots[k, :, :]
        if flag == 1:
            R = J @ R @ J
        regrot[k, :, :] = Q_mat.T @ R

    return regrot


def get_rots_mse(rots_reg, rots_ref):
    """
    Calculate MSE between the estimated orientations to reference ones.

    :param rots_reg: The estimated rotations after alignment in the form of
        a n-by-3-by-3 array.
    :param rots_ref: The reference rotations.
    :return: The MSE value between two sets of rotations.
    """
    assert (
        rots_reg.shape == rots_ref.shape
    ), "Two sets of rotations must have same dimensions."
    K = rots_reg.shape[0]

    diff = np.zeros(K)
    mse = 0
    for k in range(K):
        diff[k] = norm(rots_reg[k, :, :] - rots_ref[k, :, :], ord="fro")
        mse += diff[k] ** 2
    mse = mse / K
    return mse


def mean_aligned_angular_distance(rots_est, rots_gt, degree_tol=None):
    """
    Register estimates to ground truth rotations and compute the
    mean angular distance between them (in degrees).

    :param rots_est: A set of estimated rotations of size nx3x3.
    :param rots_gt: A set of ground truth rotations of size nx3x3.
    :param degree_tol: Option to assert if the mean angular distance is
        less than `degree_tol` degrees. If `None`, returns the mean
        aligned angular distance.

    :return: The mean angular distance between registered estimates
        and the ground truth (in degrees).
    """
    Q_mat, flag = register_rotations(rots_est, rots_gt)
    logger.debug(f"Registration Q_mat: {Q_mat}\nflag: {flag}")
    regrot = get_aligned_rotations(rots_est, Q_mat, flag)
    mean_ang_dist = Rotation.mean_angular_distance(regrot, rots_gt) * 180 / np.pi

    if degree_tol is not None:
        np.testing.assert_array_less(mean_ang_dist, degree_tol)

    return mean_ang_dist


def common_line_from_rots(r1, r2, ell):
    """
    Compute the common line induced by rotation matrices r1 and r2.

    :param r1: The first rotation matrix of 3-by-3 array.
    :param r2: The second rotation matrix of 3-by-3 array.
    :param ell: The total number of common lines.
    :return: The common line indices for both first and second rotations.
    """

    assert r1.dtype == r2.dtype, "Ambiguous dtypes"

    ut = np.dot(r2, r1.T)
    alpha_ij = np.arctan2(ut[2, 0], -ut[2, 1]) + np.pi
    alpha_ji = np.arctan2(-ut[0, 2], ut[1, 2]) + np.pi

    ell_ij = alpha_ij * ell / (2 * np.pi)
    ell_ji = alpha_ji * ell / (2 * np.pi)

    ell_ij = int(np.mod(np.round(ell_ij), ell))
    ell_ji = int(np.mod(np.round(ell_ji), ell))

    return ell_ij, ell_ji


def rots_to_clmatrix(rots, n_theta):
    """
    Compute the common lines matrix induced by all pairs of rotation
    matrices, `rots`, provided.

    :param rots: n_rotsx3x3 array of rotation matrices.
    :param n_theta: Number of theta values for common lines indices.

    :return: n_rots x n_rots common lines matrix.
    """
    n_rots = len(rots)
    cl_matrix = -np.ones((n_rots, n_rots))
    for i in range(n_rots):
        Ri = rots[i]
        Ri3 = Ri[:, 2]
        for j in range(i + 1, n_rots):
            Rj = rots[j]
            Rj3 = Rj[:, 2]
            common_axis = np.cross(Ri3, Rj3) / np.linalg.norm(np.cross(Ri3, Rj3))
            xij = Ri.T @ common_axis
            xji = Rj.T @ common_axis
            theta_ij = np.rad2deg(np.arctan2(xij[1], xij[0])) % 360
            theta_ji = np.rad2deg(np.arctan2(xji[1], xji[0])) % 360

            if theta_ij > 180:
                theta_ij -= 180
                theta_ji -= 180

            cl_matrix[i, j] = round(theta_ij * n_theta / 360)
            cl_matrix[j, i] = round((theta_ji % 360) * n_theta / 360)

    return cl_matrix


def crop_pad_2d(im, size, fill_value=0):
    """
    Crop/pads `im` according to `size`.

    Padding will use `fill_value`.
    Return's host/GPU array based on `im`.

    :param im: A >=2-dimensional numpy array
    :param size: Integer size of cropped/padded output
    :return: Array of shape (..., size, size)
    """

    im_y, im_x = im.shape[-2:]
    # shift terms
    start_x = math.floor(im_x / 2) - math.floor(size / 2)
    start_y = math.floor(im_y / 2) - math.floor(size / 2)

    # cropping
    if size <= min(im_y, im_x):
        return im[..., start_y : start_y + size, start_x : start_x + size]
    # padding
    elif size >= max(im_y, im_x):
        # Determine shape
        shape = list(im.shape[:-2])
        shape.extend([size, size])

        # Ensure that we return the same dtype as the input
        _full = np.full  # Default to numpy array
        if isinstance(im, xp.ndarray):
            # Use cupy when `im` _and_ xp are cupy ndarray
            # Avoids having to handle when cupy is not installed
            _full = xp.full

        to_return = _full(shape, fill_value, dtype=im.dtype)

        # when padding, start_x and start_y are negative since size is larger
        # than im_x and im_y; the below line calculates where the original image
        # is placed in relation to the (now-larger) box size
        to_return[..., -start_y : im_y - start_y, -start_x : im_x - start_x] = im
        return to_return
    else:
        # target size is between mat_x and mat_y
        raise ValueError(
            "Cannot crop and pad Image at the same time."
            "If this is really what you intended,"
            " make two seperate calls for cropping and padding."
        )


def crop_pad_3d(vol, size, fill_value=0):
    """
    Crop/pads `vol` according to `size`.

    Padding will use `fill_value`.
    Return's host/GPU array based on `vol`.

    :param vol: A >=3-dimensional numpy array
    :param size: Integer size of cropped/padded output
    :return: Array of shape (..., size, size, size)
    """

    vol_z, vol_y, vol_x = vol.shape[-3:]
    # shift terms
    start_z = math.floor(vol_z / 2) - math.floor(size / 2)
    start_y = math.floor(vol_y / 2) - math.floor(size / 2)
    start_x = math.floor(vol_x / 2) - math.floor(size / 2)

    # cropping
    if size <= min(vol_z, vol_y, vol_x):
        return vol[
            ...,
            start_z : start_z + size,
            start_y : start_y + size,
            start_x : start_x + size,
        ]
    # padding
    elif size >= max(vol_z, vol_y, vol_x):
        # Determine shape
        shape = list(vol.shape[:-3])
        shape.extend([size, size, size])

        # Ensure that we return the same dtype as the input
        _full = np.full  # Default to numpy array
        if isinstance(vol, xp.ndarray):
            # Use cupy when `vol` _and_ xp are cupy ndarray
            # Avoids having to handle when cupy is not installed
            _full = xp.full

        to_return = _full(shape, fill_value, dtype=vol.dtype)

        to_return[
            ...,
            -start_z : vol_z - start_z,
            -start_y : vol_y - start_y,
            -start_x : vol_x - start_x,
        ] = vol
        return to_return
    else:
        # target size is between min and max of (vol_x, vol_y, vol_z)
        raise ValueError(
            "Cannot crop and pad Volume at the same time."
            "If this is really what you intended,"
            " make two seperate calls for cropping and padding."
        )



import logging
from functools import partial

import numpy as np
from scipy.linalg import norm
from scipy.sparse.linalg import LinearOperator

from aspire import config
from aspire.basis import Coef
from aspire.nufft import anufft
from aspire.numeric import fft
from aspire.numeric.scipy import cg
from aspire.operators import evaluate_src_filters_on_grid
from aspire.reconstruction import Estimator, FourierKernel, FourierKernelMatrix
from aspire.volume import Volume, rotated_grids

logger = logging.getLogger(__name__)


class WeightedVolumesEstimator(Estimator):
    def __init__(self, weights, *args, **kwargs):
        """
        Weighted mean volume estimation.

        This class holds the `FourierKernelMatrix`, stored as a r x r
        matrix of volumes.  The problem being solved here is the
        minimization given by eq. (14) in the paper, rewritten as the
        normal equations in eq. (20) and more compactly in eq. (23).

        Convolution with each of these kernels is equivalent to
        performing a projection/backprojection on a volume, with the
        appropriate amplitude modifiers and CTF, and also a weighting
        term; the r^2 volumes are each of pairwise products between
        the weighting vectors given by the columns of wts.

        Note that this is a non-centered Fourier transform, so the
        zero frequency is found at index 0.

        Formulas and conventions used for Volume estimation are
        described in:

        “Structural Variability from Noisy Tomographic Projections.”
        Andén, Joakim, and Amit Singer.
        SIAM journal on imaging sciences vol. 11,2 (2018): 1441-1492.
        doi:10.1137/17M1153509

        "Cryo-EM reconstruction of continuous heterogeneity by Laplacian spectral volumes"
        Amit Moscovich, Amit Halevi, Joakim Andén and Amit Singer
        Inverse Problems, Volume 36, Number 2, 2020 IOP Publishing Ltd
        Special Issue on Cryo-Electron Microscopy and Inverse Problems
        https://doi.org/10.1088/1361-6420/ab4f55

        :param weights: Matrix of weights, n x r.
        """

        self.weights = weights
        self.r = self.weights.shape[1]
        super().__init__(*args, **kwargs)
        assert self.src.n == self.weights.shape[0]

    def __getattr__(self, name):
        if name == "precond_kernel":
            if self.preconditioner == "circulant":
                self.precond_kernel = FourierKernelMatrix(
                    1.0 / self.kernel.circularize()
                )
            else:
                if self.preconditioner and (
                    self.preconditioner.lower() not in ("none")
                ):
                    logger.warning(
                        f"Preconditioner {self.preconditioner} is not implemented, resetting to default of None."
                    )
                self.precond_kernel = None
            return self.precond_kernel

        else:
            #  raise AttributeError(name)
            return super().__getattr__(name)

    def compute_kernel(self):
        """
        Compute and return FourierKernelMatrix instance.
        """
        return FourierKernelMatrix(self._compute_kernel())

    def _compute_kernel(self):
        """
        :return: r x r matrix of volumes, shaped (r, r, 2L, 2L, 2L).
        """

        _2L = 2 * self.src.L
        # Note, because we're iteratively summing it is critical we zero this array.
        kernel = np.zeros((self.r, self.r, _2L, _2L, _2L), dtype=self.dtype)

        # Handle symmetry boosting.
        sym_rots = np.eye(3, dtype=self.dtype)[None]
        if self.boost:
            sym_rots = self.src.symmetry_group.matrices

        for i in range(0, self.src.n, self.batch_size):
            _range = np.arange(i, min(self.src.n, i + self.batch_size), dtype=int)
            sq_filters_f = evaluate_src_filters_on_grid(self.src, _range) ** 2
            amplitudes_sq = (self.src.amplitudes[_range] ** 2).astype(
                self.dtype, copy=False
            )

            for k in range(self.r):
                for j in range(k + 1):
                    weights = sq_filters_f * amplitudes_sq

                    if self.src.L % 2 == 0:
                        weights[0, :, :] = 0
                        weights[:, 0, :] = 0

                    weights *= (
                        self.weights[_range, j] * self.weights[_range, k]
                    ).reshape(1, 1, len(_range))

                    weights = np.transpose(weights, (2, 0, 1)).flatten()

                    # Apply boosting.
                    batch_kernel = np.zeros((_2L, _2L, _2L), dtype=self.dtype)
                    for sym_rot in sym_rots:
                        rotations = sym_rot @ self.src.rotations[_range]
                        pts_rot = rotated_grids(self.src.L, rotations)
                        pts_rot = pts_rot.reshape((3, -1))

                        batch_kernel += (
                            1
                            / (self.r * self.src.L**4)
                            * anufft(weights, pts_rot, (_2L, _2L, _2L), real=True)
                        )

                    kernel[k, j] += batch_kernel

                    # r x r symmetric
                    # accumulate batch entries of kernel[k,j] to kernel[j,k]
                    if j != k:
                        kernel[j, k] += batch_kernel

        kermat_f = np.zeros((self.r, self.r, _2L, _2L, _2L))
        logger.info("Computing non-centered Fourier Transform Kernel Mat")
        for k in range(self.r):
            for j in range(self.r):
                # Ensure symmetric kernel
                kernel[k, j, 0, :, :] = 0
                kernel[k, j, :, 0, :] = 0
                kernel[k, j, :, :, 0] = 0

                kernel[k, j] = fft.mdim_ifftshift(kernel[k, j], range(0, 3))
                kernel_f = fft.fftn(kernel[k, j], axes=(0, 1, 2))

                kernel_f = np.real(kernel_f)
                kermat_f[k, j] = kernel_f

        return kermat_f

    def src_backward(self):
        """
        Apply adjoint mapping to source

        :return: The adjoint mapping applied to the images, averaged over the whole dataset and expressed
            as coefficients of `basis`.
        """
        # Handle symmetry boosting.
        symmetry_group = None
        sym_order = 1
        if self.boost:
            symmetry_group = self.src.symmetry_group
            sym_order = len(symmetry_group.matrices)

        # src_vols_wt_backward
        vol_rhs = Volume(
            np.zeros((self.r, self.src.L, self.src.L, self.src.L), dtype=self.dtype)
        )

        for i in range(0, self.src.n, self.batch_size):
            for k in range(self.r):
                im = self.src.images[i : i + self.batch_size]

                batch_vol_rhs = self.src.im_backward(
                    im,
                    i,
                    self.weights[:, k],
                    symmetry_group=symmetry_group,
                ) / (self.src.n * sym_order)
                vol_rhs[k] += batch_vol_rhs.astype(self.dtype)

        res = np.sqrt(self.src.n * sym_order) * self.basis.evaluate_t(vol_rhs)
        logger.info(f"Determined weighted adjoint mappings. Shape = {res.shape}")

        return res

    def conj_grad(self, b_coef, x0=None, tol=1e-5, regularizer=0):
        count = b_coef.shape[-1]  # b_coef should be (r, basis.count)
        kernel = self.kernel

        if regularizer > 0:
            kernel += regularizer

        operator = LinearOperator(
            (self.r * count, self.r * count),
            matvec=partial(self.apply_kernel, kernel=kernel),
            dtype=self.dtype,
        )
        if self.precond_kernel is None:
            M = None
        else:
            precond_kernel = self.precond_kernel
            if regularizer > 0:
                precond_kernel += regularizer
            M = LinearOperator(
                (self.r * count, self.r * count),
                matvec=partial(self.apply_kernel, kernel=precond_kernel),
                dtype=self.dtype,
            )

        tol = tol or config.mean.cg_tol
        target_residual = tol * norm(b_coef)

        # callback setup
        self.i = 0  # iteration counter

        def cb(xk):
            self.i += 1  # increment iteration count

            logger.info(
                f"[Iter {self.i}]: Delta {norm(b_coef - self.apply_kernel(xk))} (target {target_residual})"
            )

            # Do checkpoint at `checkpoint_iterations`,
            _do_checkpoint = (
                self.checkpoint_iterations is not None
                and (self.i % self.checkpoint_iterations) == 0
            )
            # or the last iteration when `maxiter` provided.
            if self.maxiter:
                _do_checkpoint |= self.i == (self.maxiter - 1)

            # Optional checkpoint
            if _do_checkpoint:
                # Construct checkpoint path
                path = f"{self.checkpoint_prefix}_iter{self.i:04d}.npy"
                # Write out the current solution
                np.save(path, xk)
                logger.info(f"Checkpoint saved to `{path}`")

        x, info = cg(
            operator,
            b_coef.flatten(),
            x0=x0,
            M=M,
            callback=cb,
            rtol=tol,
            atol=0,
            maxiter=self.maxiter,
        )

        if info != 0:
            logger.warning(
                f"Conjugate gradient unable to converge after {info} iterations."
            )

        return x.reshape(self.r, self.basis.count)

    def apply_kernel(self, vol_coef, kernel=None):
        """
        Applies the kernel represented by convolution

        :param vol_coef: The volume to be convolved, stored in the basis coefficients.
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :return: The result of evaluating `vol_coef` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis.
        """
        if kernel is None:
            kernel = self.kernel

        assert np.size(vol_coef) == self.r * self.basis.count
        if vol_coef.ndim == 1:
            vol_coef = vol_coef.reshape(self.r, self.basis.count)

        vols_out = Volume(
            np.zeros((self.r, self.src.L, self.src.L, self.src.L), dtype=self.dtype)
        )

        vol = Coef(self.basis, vol_coef).evaluate()

        for k in range(self.r):
            for j in range(self.r):
                vols_out[k] = vols_out[k] + kernel.convolve_volume(vol[j], j, k)
                # Note this is where we would add mask_gamma

        vol_coef = self.basis.evaluate_t(vols_out)

        return vol_coef


class MeanEstimator(WeightedVolumesEstimator):
    """
    Special case of weighted mean volume estimate,
    for a single volume.
    """

    def __init__(self, src, **kwargs):
        # Note, Handle boosting by adjusting weights based on symmetric order.
        weights = np.ones((src.n, 1)) / np.sqrt(
            src.n * len(src.symmetry_group.matrices)
        )
        super().__init__(weights, src, **kwargs)

    def __getattr__(self, name):
        """
        See `Estimator.__getattr__`.
        """
        return super(WeightedVolumesEstimator, self).__getattr__(name)

    def apply_kernel(self, *args, **kwargs):
        """
        See `Estimator.apply_kernel`.
        """
        return super(WeightedVolumesEstimator, self).apply_kernel(*args, **kwargs)

    def compute_kernel(self):
        """
        Compute and return `FourierKernel` instance.
        """
        # Note for the r=1 we select and return a single kernel.
        return FourierKernel(self._compute_kernel()[0][0])



import logging
from abc import ABC, abstractmethod

from aspire.source.image import _ImageAccessor

logger = logging.getLogger(__name__)


class Denoiser(ABC):
    """
    Base class for 2D image denoisers.
    """

    def __init__(self, src):
        """
        Initialize an object for denoising 2D images from `src`.

        :param src: `ImageSource` providing noisy images.
        """

        self.src = src
        self.dtype = src.dtype
        self.n = src.n
        self._img_accessor = _ImageAccessor(self._denoise, self.n)

    @property
    def denoise(self):
        """
        Subscriptable property returning 2D images after denoising.

        See `_ImageAccessor`.
        """
        return self._img_accessor

    @abstractmethod
    def _denoise(self, indices):
        """
        Subclasses must implement a private `_denoise` method accepting `indices`.
        Subclasses handle any caching as well as denoising.

        See `_ImageAccessor`.
        """



import logging

from aspire.denoising import Denoiser
from aspire.source import ImageSource

logger = logging.getLogger(__name__)


class DenoisedSource(ImageSource):
    """
    `ImageSource` class serving denoised 2D images.
    """

    def __init__(self, src, denoiser):
        """
        Initialize a denoised `ImageSource` object from an `ImageSource`.

        :param src: Original `ImageSource` object storing noisy images
        :param denoiser: A `Denoiser` object for specifying a method for denoising
        """

        super().__init__(src.L, src.n, dtype=src.dtype, metadata=src._metadata.copy())
        # TODO, we can probably setup a reasonable default here.
        self.denoiser = denoiser
        if not isinstance(denoiser, Denoiser):
            raise TypeError("`denoiser` must be subclass of `Denoiser`")

        # Safety check src and self.denoiser.src are the same.
        # See #1020
        if src != self.denoiser.src:
            raise NotImplementedError(
                "Denoiser `src` and noisy image `src` must match."
            )

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _images(self, indices):
        """
        Internal function to return a set of images after denoising, when accessed via the
        `ImageSource.images` property.

        :param indices: The indices of images to return as a 1-D Numpy array.
        :return: an `Image` object after denoising.
        """

        # check for cached images first
        if self._cached_im is not None:
            logger.info("Loading images from cache")
            return self.generation_pipeline.forward(
                self._cached_im[indices, :, :], indices
            )

        imgs_denoised = self.denoiser.denoise[indices]

        # Finally, apply transforms to resulting Image
        return self.generation_pipeline.forward(imgs_denoised, indices)



import abc

import numpy as np
from numpy.linalg import qr

from aspire.utils import bump_3d, grid_3d
from aspire.utils.random import Random, randn
from aspire.volume import (
    CnSymmetryGroup,
    DnSymmetryGroup,
    IdentitySymmetryGroup,
    OSymmetryGroup,
    TSymmetryGroup,
    Volume,
)


class SyntheticVolumeBase(abc.ABC):
    def __init__(self, L, C, pixel_size=None, seed=None, dtype=np.float64):
        self.L = L
        self.C = C
        self.seed = seed
        self.dtype = dtype
        self.pixel_size = pixel_size

    @abc.abstractmethod
    def generate(self):
        """
        Called to generate and return synthetic volumes.

        Each concrete subclass should implement this.
        """

    def __repr__(self):
        return f"{self.__class__.__name__} {self.__dict__}"


class GaussianBlobsVolume(SyntheticVolumeBase):
    """
    A base class for all volumes which are generated with randomized 3D Gaussians.
    """

    def __init__(
        self, L, C, K=16, alpha=1, pixel_size=None, seed=None, dtype=np.float64
    ):
        """
        :param L: Resolution of the Volume(s) in pixels.
        :param C: Number of Volumes to generate.
        :param K: Number of Gaussian blobs used to construct the Volume(s).
        :param alpha: Scaling factor for variance of Gaussian blobs. Default=1.
        :param pixel_size: Optional voxel_size in angstroms.
            When provided will be saved with `map`/`mrc` metadata.
            Default of `None` will not write to file,
            but will be considered unit pixels (1) for FSC.
        :param seed: Random seed for generating random Gaussian blobs.
        :param dtype: dtype for Volume(s)
        """
        self.K = int(K)
        self.alpha = float(alpha)
        super().__init__(L=L, C=C, pixel_size=pixel_size, seed=seed, dtype=dtype)
        self._set_symmetry_group()

    @abc.abstractproperty
    def n_blobs(self):
        """
        The total number of Gaussian blobs used to generate a Volume.
        This value differs from `self.K` as it accounts for the blobs
        which have been duplicated during `_symmetrize_gaussians`.
        """

    @property
    def symmetry_group(self):
        """
        SymmetryGroup object corresponding to the symmetry of the Volume.
        """
        return self._symmetry_group

    def generate(self):
        """
        Generates a Volume object with specified symmetry that is multiplied by a bump function
        to give compact support within the unit sphere.
        """
        vol = self._gaussian_blob_vols()
        bump_mask = bump_3d(self.L, spread=5, dtype=self.dtype)
        return Volume(
            bump_mask * vol,
            symmetry_group=self.symmetry_group,
            pixel_size=self.pixel_size,
        )

    def _gaussian_blob_vols(self):
        """
        Generates a 4D array representing a stack of volumes composed of Gaussian blobs.

        :return: An ndarray containing C Gaussian blob volumes.
        """
        vols = np.zeros(shape=((self.C,) + (self.L,) * 3)).astype(self.dtype)
        with Random(self.seed):
            for c in range(self.C):
                Q, D, mu = self._gen_gaussians()
                Q_rot, D_sym, mu_rot = self._symmetrize_gaussians(Q, D, mu)
                vols[c] = self._eval_gaussians(Q_rot, D_sym, mu_rot)
        return vols

    def _gen_gaussians(self):
        """
        For K gaussians, generate random orientation (Q), mean (mu), and variance (D).

        :return: Orientations Q, Variances D, Means mu.
        """
        Q = np.zeros(shape=(self.K, 3, 3)).astype(self.dtype)
        D = np.zeros(shape=(self.K, 3, 3)).astype(self.dtype)
        mu = np.zeros(shape=(self.K, 3)).astype(self.dtype)

        for k in range(self.K):
            V = randn(3, 3).astype(self.dtype) / np.sqrt(3)
            Q[k, :, :] = qr(V)[0]
            D[k, :, :] = (
                self.alpha**2 / self.n_blobs * np.diag(np.sum(abs(V) ** 2, axis=0))
            )
            mu[k, :] = 0.5 * randn(3) / np.sqrt(3)

        return Q, D, mu

    def _symmetrize_gaussians(self, Q, D, mu):
        """
        Called to add symmetry to Volumes by generating for each Gaussian blob duplicates in symmetric positions.
        """
        rots = self.symmetry_group.matrices

        Q_rot = np.zeros(shape=(self.n_blobs, 3, 3)).astype(self.dtype)
        D_sym = np.zeros(shape=(self.n_blobs, 3, 3)).astype(self.dtype)
        mu_rot = np.zeros(shape=(self.n_blobs, 3)).astype(self.dtype)
        idx = 0

        for rot in rots:
            for k in range(self.K):
                Q_rot[idx] = rot.T @ Q[k]
                D_sym[idx] = D[k]
                mu_rot[idx] = rot.T @ mu[k]
                idx += 1
        return Q_rot, D_sym, mu_rot

    def _eval_gaussians(self, Q, D, mu):
        """
        Evaluate Gaussian blobs over a 3D grid with centers, mu, orientations, Q, and variances, D.

        :param Q: A stack of size (n_blobs) x 3 x 3 of rotation matrices,
            determining the orientation of each blob.
        :param D: A stack of size (n_blobs) x 3 x 3 diagonal matrices,
            whose diagonal entries are the variances of each blob.
        :param mu: An array of size (n_blobs) x 3 containing the centers for each blob.

        :return: An L x L x L array.
        """
        g = grid_3d(self.L, indexing="zyx", dtype=self.dtype)
        coords = np.array(
            [g["x"].flatten(), g["y"].flatten(), g["z"].flatten()], dtype=self.dtype
        )

        n_blobs = Q.shape[0]
        vol = np.zeros(shape=(1, coords.shape[-1])).astype(self.dtype)

        for k in range(n_blobs):
            coords_k = coords - mu[k, :, np.newaxis]
            coords_k = (
                Q[k].T @ coords_k * np.sqrt(1 / np.diag(D[k, :, :]))[:, np.newaxis]
            )

            vol += np.exp(-0.5 * np.sum(np.abs(coords_k) ** 2, axis=0))

        vol = np.reshape(vol, g["x"].shape)

        return vol


class CnSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with cyclically symmetric volumes constructed of random 3D Gaussian blobs.
    """

    def __init__(
        self, L, C, order, K=16, alpha=1, pixel_size=None, seed=None, dtype=np.float64
    ):
        """
        :param L: Resolution of the Volume(s) in pixels.
        :param C: Number of Volumes to generate.
        :param order: An integer representing the cyclic order of the Volume(s).
        :param K: Number of Gaussian blobs used to construct the Volume(s).
        :param pixel_size: Optional voxel_size in angstroms.
            When provided will be saved with `map`/`mrc` metadata.
            Default of `None` will not write to file,
            but will be considered unit pixels (1) for FSC.
        :param seed: Random seed for generating random Gaussian blobs.
        :param dtype: dtype for Volume(s)
        """
        self.order = int(order)
        self._check_order()
        super().__init__(
            L=L, C=C, K=K, alpha=alpha, pixel_size=pixel_size, seed=seed, dtype=dtype
        )

    def _check_order(self):
        if self.order < 2:
            raise ValueError(
                f"For a {self.__class__.__name__} the cyclic order must be greater than 1. Provided order was {self.order}"
            )

    def _set_symmetry_group(self):
        self._symmetry_group = CnSymmetryGroup(order=self.order, dtype=self.dtype)

    @property
    def n_blobs(self):
        return self.order * self.K


class DnSymmetricVolume(CnSymmetricVolume):
    """
    A Volume object with n-fold dihedral symmetry constructed of random 3D Gaussian blobs.
    """

    def _set_symmetry_group(self):
        self._symmetry_group = DnSymmetryGroup(order=self.order, dtype=self.dtype)

    @property
    def n_blobs(self):
        return 2 * self.order * self.K


class TSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with tetrahedral symmetry constructed of random 3D Gaussian blobs.
    """

    def _set_symmetry_group(self):
        self._symmetry_group = TSymmetryGroup(dtype=self.dtype)

    @property
    def n_blobs(self):
        return 12 * self.K


class OSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with octahedral symmetry constructed of random 3D Gaussian blobs.
    """

    def _set_symmetry_group(self):
        self._symmetry_group = OSymmetryGroup(dtype=self.dtype)

    @property
    def n_blobs(self):
        return 24 * self.K


class AsymmetricVolume(CnSymmetricVolume):
    """
    An asymmetric Volume constructed of random 3D Gaussian blobs with compact support in the unit sphere.
    """

    def __init__(self, L, C, K=64, pixel_size=None, seed=None, dtype=np.float64):
        super().__init__(
            L=L, C=C, K=K, order=1, pixel_size=pixel_size, seed=seed, dtype=dtype
        )

    def _check_order(self):
        if self.order != 1:
            raise ValueError(
                f"An {self.__class__.__name__} must have order=1. Provided order was {self.order}"
            )

    def _set_symmetry_group(self):
        self._symmetry_group = IdentitySymmetryGroup(dtype=self.dtype)

    def _symmetrize_gaussians(self, Q, D, mu):
        return Q, D, mu


class LegacyVolume(AsymmetricVolume):
    """
    An asymmetric Volume object used for testing of legacy code.
    """

    def __init__(self, L, C=2, K=16, pixel_size=None, seed=0, dtype=np.float64):
        super().__init__(L=L, C=C, K=K, pixel_size=pixel_size, seed=seed, dtype=dtype)

    def generate(self):
        """
        Generates an asymmetric volume composed of random 3D Gaussian blobs.
        """
        vols = self._gaussian_blob_vols()

        # Swap axes to retain Legacy xyz-indexing.
        vols = np.swapaxes(vols, 1, 3)

        return Volume(vols, pixel_size=self.pixel_size)
